# Unit test check_preprocess_exog_multiseries
# ==============================================================================
import re
import pytest
import numpy as np
import pandas as pd
from skforecast.exceptions import MissingExogWarning
from skforecast.utils import check_preprocess_series
from skforecast.utils import check_preprocess_exog_multiseries
from skforecast.ForecasterAutoregMultiSeries.tests.fixtures_ForecasterAutoregMultiSeries import series
from skforecast.ForecasterAutoregMultiSeries.tests.fixtures_ForecasterAutoregMultiSeries import series_2
from skforecast.ForecasterAutoregMultiSeries.tests.fixtures_ForecasterAutoregMultiSeries import series_as_dict
from skforecast.ForecasterAutoregMultiSeries.tests.fixtures_ForecasterAutoregMultiSeries import exog
from skforecast.ForecasterAutoregMultiSeries.tests.fixtures_ForecasterAutoregMultiSeries import exog_as_dict
from skforecast.ForecasterAutoregMultiSeries.tests.fixtures_ForecasterAutoregMultiSeries import exog_as_dict_datetime
from skforecast.ForecasterAutoregMultiSeries.tests.fixtures_ForecasterAutoregMultiSeries import exog_predict


def test_TypeError_check_preprocess_exog_multiseries_when_exog_is_not_valid_type():
    """
    Test TypeError is raised when exog is not a pandas Series, DataFrame or dict.
    """
    _, series_indexes = check_preprocess_series(series=series)

    not_valid_exog = 'not_valid_exog'

    err_msg = re.escape(
        ("`exog` must be a pandas Series, DataFrame, dictionary of pandas "
         "Series/DataFrames or None. Got <class 'str'>.")
    )
    with pytest.raises(TypeError, match = err_msg):
        check_preprocess_exog_multiseries(
            input_series_is_dict = False,
            series_indexes       = series_indexes,
            series_col_names     = ['1', '2'],
            exog                 = not_valid_exog,
            exog_dict            = {'1': None, '2': None}
        )


@pytest.mark.parametrize("not_valid_exog", 
                         [pd.Series([1, 2, 3], name='exog1'),
                          pd.DataFrame([1, 2, 3], columns=['exog1'])], 
                         ids=lambda exog: f'exog: {type(exog)}')
def test_TypeError_check_preprocess_exog_multiseries_when_exog_is_not_dict_and_series_is_dict(not_valid_exog):
    """
    Test TypeError is raised when exog is a pandas Series or DataFrame and
    input series is a dict (input_series_is_dict = True).
    """
    _, series_indexes = check_preprocess_series(series=series_as_dict)

    err_msg = re.escape(
        (f"`exog` must be a dict of DataFrames or Series if "
         f"`series` is a dict. Got {type(not_valid_exog)}.")
    )
    with pytest.raises(TypeError, match = err_msg):
        check_preprocess_exog_multiseries(
            input_series_is_dict = True,
            series_indexes       = series_indexes,
            series_col_names     = ['l1', 'l2'],
            exog                 = not_valid_exog,
            exog_dict            = {'l1': None, 'l2': None}
        )


@pytest.mark.parametrize("not_valid_exog", 
                         [exog['exog_1'].iloc[:30].copy(),
                          exog.iloc[:30].copy()], 
                         ids=lambda exog: f'exog: {type(exog)}')
def test_ValueError_check_preprocess_exog_multiseries_when_exog_pandas_with_different_length_from_series(not_valid_exog):
    """
    Test ValueError is raised when exog is a pandas Series or DataFrame with a
    different length from input series (series_indexes).
    """
    _, series_indexes = check_preprocess_series(series=series)

    err_msg = re.escape(
        (f"`exog` must have same number of samples as `series`. "
         f"length `exog`: ({len(not_valid_exog)}), length `series`: ({len(series)})")
    )
    with pytest.raises(ValueError, match = err_msg):
        check_preprocess_exog_multiseries(
            input_series_is_dict = False,
            series_indexes       = series_indexes,
            series_col_names     = ['1', '2'],
            exog                 = not_valid_exog,
            exog_dict            = {'1': None, '2': None}
        )


def test_ValueError_check_preprocess_exog_multiseries_when_exog_pandas_with_different_index_from_series():
    """
    Test ValueError is raised when exog is a pandas Series or DataFrame with a
    different index from input series (series_indexes).
    """
    _, series_indexes = check_preprocess_series(series=series)

    err_msg = re.escape(
        ("Different index for `series` and `exog`. They must be equal "
         "to ensure the correct alignment of values.")
    )
    with pytest.raises(ValueError, match = err_msg):
        check_preprocess_exog_multiseries(
            input_series_is_dict = False,
            series_indexes       = series_indexes,
            series_col_names     = ['1', '2'],
            exog                 = exog_predict,
            exog_dict            = {'1': None, '2': None}
        )
    

def test_check_preprocess_exog_multiseries_when_series_is_pandas_DataFrame():
    """
    Test check_preprocess_exog_multiseries when `series` is a pandas DataFrame.
    """
    _, series_indexes = check_preprocess_series(series=series)

    exog_dict, exog_col_names = check_preprocess_exog_multiseries(
                                    input_series_is_dict = False,
                                    series_indexes       = series_indexes,
                                    series_col_names     = ['1', '2'],
                                    exog                 = exog,
                                    exog_dict            = {'1': None, '2': None}
                                )

    expected_exog_dict = {
        '1': exog,
        '2': exog
    }
    expected_exog_col_names = ['exog_1', 'exog_2']

    assert isinstance(exog_dict, dict)
    assert list(exog_dict.keys()) == ['1', '2']
    for k in exog_dict:
        pd.testing.assert_frame_equal(exog_dict[k], expected_exog_dict[k])
        pd.testing.assert_index_equal(exog_dict[k].index, series_indexes['1'])
    
    assert len(set(exog_col_names) - set(expected_exog_col_names)) == 0
    

def test_check_preprocess_exog_multiseries_when_series_is_pandas_DataFrame_with_RangeIndex_not_starting_0():
    """
    Test check_preprocess_exog_multiseries when `series` is a pandas DataFrame with 
    a RangeIndex that doesn't start at 0.
    """
    series_range = series.copy()
    series_range.index = pd.RangeIndex(start=3, stop=3+len(series), step=1)
    _, series_indexes = check_preprocess_series(series=series_range)
    
    exog_range = exog.copy()
    exog_range.index = pd.RangeIndex(start=3, stop=3+len(exog_range), step=1)

    exog_dict, exog_col_names = check_preprocess_exog_multiseries(
                                    input_series_is_dict = False,
                                    series_indexes       = series_indexes,
                                    series_col_names     = ['1', '2'],
                                    exog                 = exog_range,
                                    exog_dict            = {'1': None, '2': None}
                                )

    expected_exog_dict = {
        '1': exog_range,
        '2': exog_range
    }
    expected_exog_col_names = ['exog_1', 'exog_2']

    assert isinstance(exog_dict, dict)
    assert list(exog_dict.keys()) == ['1', '2']
    for k in exog_dict:
        pd.testing.assert_frame_equal(exog_dict[k], expected_exog_dict[k])
        pd.testing.assert_index_equal(exog_dict[k].index, series_indexes['1'])
    
    assert len(set(exog_col_names) - set(expected_exog_col_names)) == 0
    

def test_output_check_preprocess_exog_multiseries_when_series_is_pandas_DataFrame_with_DatetimeIndex():
    """
    Test check_preprocess_exog_multiseries when `series` is a pandas DataFrame with 
    a DatetimeIndex.
    """
    series_datetime = series.copy()
    series_datetime.index = pd.date_range(start='2000-01-01', periods=len(series), freq='D')
    _, series_indexes = check_preprocess_series(series=series_datetime)

    exog_datetime = exog.copy()
    exog_datetime.index = pd.date_range(
        start='2000-01-01', periods=len(exog_datetime), freq='D'
    )

    exog_dict, exog_col_names = check_preprocess_exog_multiseries(
                                    input_series_is_dict = False,
                                    series_indexes       = series_indexes,
                                    series_col_names     = ['1', '2'],
                                    exog                 = exog_datetime,
                                    exog_dict            = {'1': None, '2': None}
                                )

    expected_exog_dict = {
        '1': exog_datetime,
        '2': exog_datetime
    }
    expected_exog_col_names = ['exog_1', 'exog_2']

    assert isinstance(exog_dict, dict)
    assert list(exog_dict.keys()) == ['1', '2']
    for k in exog_dict:
        pd.testing.assert_frame_equal(exog_dict[k], expected_exog_dict[k])
        pd.testing.assert_index_equal(exog_dict[k].index, series_indexes['1'])
    
    assert len(set(exog_col_names) - set(expected_exog_col_names)) == 0


def test_TypeError_check_preprocess_exog_multiseries_when_exog_is_dict_with_no_pandas_Series_or_DataFrame():
    """
    Test TypeError is raised when exog is dict not containing 
    a pandas Series or DataFrame.
    """
    _, series_indexes = check_preprocess_series(series=series_2)

    not_valid_exog = {
        'l1': exog_as_dict_datetime['l1'].copy(),
        'l2': np.array([1, 2, 3]),
        'l3': None
    }

    err_msg = re.escape(
        ("If `exog` is a dictionary, all exog must be a named pandas "
         "Series, a pandas DataFrame or None. Review exog: ['l2']")
    )
    with pytest.raises(TypeError, match = err_msg):
        check_preprocess_exog_multiseries(
            input_series_is_dict = False,
            series_indexes       = series_indexes,
            series_col_names     = ['l1', 'l2'],
            exog                 = not_valid_exog,
            exog_dict            = {'l1': None, 'l2': None}
        )


def test_MissingValuesWarning_check_preprocess_exog_multiseries_when_exog_is_dict_without_all_series():
    """
    Test MissingValuesWarning is issues when exog is a dict without all the 
    series as keys.
    """
    _, series_indexes = check_preprocess_series(series=series_2)

    incomplete_exog = {
        'l1': exog_as_dict_datetime['l1'].copy(),
        'l2': exog_as_dict_datetime['l2'].copy()
    }
    incomplete_exog.pop('l1')

    warn_msg = re.escape(
        ("{'l1'} not present in `exog`. All values "
         "of the exogenous variables for these series will be NaN.")
    )
    with pytest.warns(MissingExogWarning, match = warn_msg):
        check_preprocess_exog_multiseries(
            input_series_is_dict = False,
            series_indexes       = series_indexes,
            series_col_names     = ['l1', 'l2'],
            exog                 = incomplete_exog,
            exog_dict            = {'l1': None, 'l2': None}
        )


def test_ValueError_check_preprocess_exog_multiseries_when_exog_dict_with_different_length_from_series():
    """
    Test ValueError is raised when exog is a dict with different length from 
    input series as pandas DataFrame (series_indexes).
    """
    _, series_indexes = check_preprocess_series(series=series_2)

    not_valid_exog = {
        'l1': exog_as_dict_datetime['l1'].copy(),
        'l2': exog_as_dict_datetime['l2'].copy()
    }
    not_valid_exog['l2'] = not_valid_exog['l2'].iloc[:30]

    err_msg = re.escape(
        ("`exog` for series 'l2' must have same number of samples as `series`. "
         "length `exog`: (30), length `series`: (50)")
    )
    with pytest.raises(ValueError, match = err_msg):
        check_preprocess_exog_multiseries(
            input_series_is_dict = False,
            series_indexes       = series_indexes,
            series_col_names     = ['l1', 'l2'],
            exog                 = not_valid_exog,
            exog_dict            = {'l1': None, 'l2': None}
        )


def test_ValueError_check_preprocess_exog_multiseries_when_exog_dict_with_different_index_from_series():
    """
    Test ValueError is raised when exog is a dict with different index from 
    input series as pandas DataFrame (series_indexes).
    """
    _, series_indexes = check_preprocess_series(series=series)

    not_valid_exog = {
        '1': exog.copy(),
        '2': exog_predict['exog_1'].copy()
    }

    err_msg = re.escape(
        ("Different index for series '2' and its exog. "
         "When `series` is a pandas DataFrame, they must "
         "be equal to ensure the correct alignment of values.")
    )
    with pytest.raises(ValueError, match = err_msg):
        check_preprocess_exog_multiseries(
            input_series_is_dict = False,
            series_indexes       = series_indexes,
            series_col_names     = ['1', '2'],
            exog                 = not_valid_exog,
            exog_dict            = {'1': None, '2': None}
        )


def test_TypeError_check_preprocess_exog_multiseries_when_exog_dict_with_different_index_from_series():
    """
    Test TypeError is raised when exog is a dict with not DatetimeIndex index 
    when input series is a dict (input_series_is_dict=True).
    """
    _, series_indexes = check_preprocess_series(series=series_as_dict)

    err_msg = re.escape(
        ("All exog must have a Pandas DatetimeIndex as index with the "
         "same frequency. Check exog for series: ['l1', 'l2']")
    )
    with pytest.raises(TypeError, match = err_msg):
        check_preprocess_exog_multiseries(
            input_series_is_dict = True,
            series_indexes       = series_indexes,
            series_col_names     = ['l1', 'l2'],
            exog                 = exog_as_dict,
            exog_dict            = {'l1': None, 'l2': None}
        )


def test_TypeError_check_preprocess_exog_multiseries_when_exog_dict_with_different_dtypes_same_column():
    """
    Test TypeError is raised when exog is a dict with different dtypes for the 
    same column.
    """
    _, series_indexes = check_preprocess_series(series=series)

    not_valid_exog = {
        '1': exog.copy(),
        '2': exog['exog_1'].astype(str).copy()
    }

    err_msg = re.escape(
        ("Column 'exog_1' has different dtypes in different exog "
         "DataFrames or Series.")
    )
    with pytest.raises(TypeError, match = err_msg):
        check_preprocess_exog_multiseries(
            input_series_is_dict = False,
            series_indexes       = series_indexes,
            series_col_names     = ['1', '2'],
            exog                 = not_valid_exog,
            exog_dict            = {'1': None, '2': None}
        )


def test_ValueError_check_preprocess_exog_multiseries_when_exog_has_columns_named_as_series():
    """
    Test ValueError is raised when exog has columns named as the series.
    """
    _, series_indexes = check_preprocess_series(series=series)

    duplicate_exog = exog['exog_1'].copy()
    duplicate_exog.name = '1'

    not_valid_exog = {
        '1': duplicate_exog
    }

    err_msg = re.escape(
        ("`exog` cannot contain a column named the same as one of the "
         "series (column names of series).\n"
         "    `series` columns : ['1', '2'].\n"
         "    `exog`   columns : ['1'].")
    )
    with pytest.raises(ValueError, match = err_msg):
        check_preprocess_exog_multiseries(
            input_series_is_dict = False,
            series_indexes       = series_indexes,
            series_col_names     = ['1', '2'],
            exog                 = not_valid_exog,
            exog_dict            = {'1': None, '2': None}
        )


def test_output_check_preprocess_exog_multiseries_when_series_is_DataFrame_and_exog_pandas_Series():
    """
    Test check_preprocess_exog_multiseries when `series` is a pandas DataFrame 
    and `exog` is a pandas Series.
    """
    series_dict, series_indexes = check_preprocess_series(series=series)

    exog_dict, exog_col_names = check_preprocess_exog_multiseries(
                                    input_series_is_dict = False,
                                    series_indexes       = series_indexes,
                                    series_col_names     = ['1', '2'],
                                    exog                 = exog['exog_1'],
                                    exog_dict            = {'1': None, '2': None}
                                )

    expected_exog_dict = {
        '1': pd.DataFrame(
                 data    = exog['exog_1'].to_numpy(),
                 index   = pd.RangeIndex(start=0, stop=50),
                 columns = ['exog_1']
             ),
        '2': pd.DataFrame(
                 data    = exog['exog_1'].to_numpy(),
                 index   = pd.RangeIndex(start=0, stop=50),
                 columns = ['exog_1']
             ),
    }
    expected_exog_col_names = ['exog_1']
    
    assert isinstance(exog_dict, dict)
    assert list(exog_dict.keys()) == ['1', '2']
    for k in exog_dict:
        assert isinstance(exog_dict[k], pd.DataFrame)
        pd.testing.assert_frame_equal(exog_dict[k], expected_exog_dict[k])
        pd.testing.assert_index_equal(exog_dict[k].index, series_dict[k].index)

    assert exog_col_names == expected_exog_col_names


def test_output_check_preprocess_exog_multiseries_when_series_is_DataFrame_and_exog_pandas_DataFrame():
    """
    Test check_preprocess_exog_multiseries when `series` is a pandas DataFrame 
    and `exog` is a pandas DataFrame.
    """
    series_dict, series_indexes = check_preprocess_series(series=series)

    exog_dict, exog_col_names = check_preprocess_exog_multiseries(
                                    input_series_is_dict = False,
                                    series_indexes       = series_indexes,
                                    series_col_names     = ['1', '2'],
                                    exog                 = exog,
                                    exog_dict            = {'1': None, '2': None}
                                )

    expected_exog_dict = {
        '1': pd.DataFrame(
                 data    = exog.to_numpy(),
                 index   = pd.RangeIndex(start=0, stop=50),
                 columns = ['exog_1', 'exog_2']
             ).astype({'exog_1': float}),
        '2': pd.DataFrame(
                 data    = exog.to_numpy(),
                 index   = pd.RangeIndex(start=0, stop=50),
                 columns = ['exog_1', 'exog_2']
             ).astype({'exog_1': float}),
    }
    expected_exog_col_names = ['exog_1', 'exog_2']
    
    assert isinstance(exog_dict, dict)
    assert list(exog_dict.keys()) == ['1', '2']
    for k in exog_dict:
        assert isinstance(exog_dict[k], pd.DataFrame)
        pd.testing.assert_frame_equal(exog_dict[k], expected_exog_dict[k])
        pd.testing.assert_index_equal(exog_dict[k].index, series_dict[k].index)

    assert len(set(exog_col_names) - set(expected_exog_col_names)) == 0


def test_output_check_preprocess_exog_multiseries_when_series_is_DataFrame_and_exog_dict():
    """
    Test check_preprocess_exog_multiseries when `series` is a pandas DataFrame 
    and `exog` is a dict.
    """
    series_dict, series_indexes = check_preprocess_series(series=series_2)

    exog_dict, exog_col_names = check_preprocess_exog_multiseries(
                                    input_series_is_dict = False,
                                    series_indexes       = series_indexes,
                                    series_col_names     = ['l1', 'l2'],
                                    exog                 = exog_as_dict_datetime,
                                    exog_dict            = {'l1': None, 'l2': None}
                                )

    expected_exog_dict = {
        'l1': pd.DataFrame(
                  data    = exog_as_dict_datetime['l1'].to_numpy(),
                  index   = pd.date_range(start='2000-01-01', periods=len(exog), freq='D'),
                  columns = ['exog_1', 'exog_2']
              ).astype({'exog_1': float}),
        'l2': pd.DataFrame(
                  data    = exog_as_dict_datetime['l2'].to_numpy(),
                  index   = pd.date_range(start='2000-01-01', periods=len(exog), freq='D'),
                  columns = ['exog_1']
              ).astype({'exog_1': float}),
    }
    expected_exog_col_names = ['exog_1', 'exog_2']
    
    assert isinstance(exog_dict, dict)
    assert list(exog_dict.keys()) == ['l1', 'l2']
    for k in exog_dict:
        assert isinstance(exog_dict[k], pd.DataFrame)
        pd.testing.assert_frame_equal(exog_dict[k], expected_exog_dict[k])
        pd.testing.assert_index_equal(exog_dict[k].index, series_dict[k].index)

    assert len(set(exog_col_names) - set(expected_exog_col_names)) == 0


def test_output_check_preprocess_exog_multiseries_when_series_is_dict_and_exog_dict():
    """
    Test check_preprocess_exog_multiseries when `series` is a dict (input_series_is_dict = True)
    and `exog` is a dict.
    """
    series_dict, series_indexes = check_preprocess_series(series=series_as_dict)

    exog_dict, exog_col_names = check_preprocess_exog_multiseries(
                                    input_series_is_dict = True,
                                    series_indexes       = series_indexes,
                                    series_col_names     = ['l1', 'l2'],
                                    exog                 = exog_as_dict_datetime,
                                    exog_dict            = {'l1': None, 'l2': None}
                                )

    expected_exog_dict = {
        'l1': pd.DataFrame(
                  data    = exog_as_dict_datetime['l1'].to_numpy(),
                  index   = pd.date_range(start='2000-01-01', periods=len(exog), freq='D'),
                  columns = ['exog_1', 'exog_2']
              ).astype({'exog_1': float}),
        'l2': pd.DataFrame(
                  data    = exog_as_dict_datetime['l2'].to_numpy(),
                  index   = pd.date_range(start='2000-01-01', periods=len(exog), freq='D'),
                  columns = ['exog_1']
              ).astype({'exog_1': float}),
    }
    expected_exog_col_names = ['exog_1', 'exog_2']
    
    assert isinstance(exog_dict, dict)
    assert list(exog_dict.keys()) == ['l1', 'l2']
    for k in exog_dict:
        assert isinstance(exog_dict[k], pd.DataFrame)
        pd.testing.assert_frame_equal(exog_dict[k], expected_exog_dict[k])
        pd.testing.assert_index_equal(exog_dict[k].index, series_dict[k].index)

    assert len(set(exog_col_names) - set(expected_exog_col_names)) == 0


def test_output_check_preprocess_exog_multiseries_when_series_is_dict_and_exog_dict_with_different_lengths():
    """
    Test check_preprocess_exog_multiseries when `series` is a dict (input_series_is_dict = True)
    and `exog` is a dict with series of different lengths and exog None.
    """
    series_dict_3 = {
        'l1': series_as_dict['l1'].copy(),
        'l2': series_as_dict['l2'].copy(),
        'l3': series_as_dict['l1'].copy()
    }
    series_dict, series_indexes = check_preprocess_series(series=series_dict_3)

    exog_test = {
        'l1': exog_as_dict_datetime['l1'].copy(),
        'l2': exog_as_dict_datetime['l2'].copy(),
        'l3': None
    }
    exog_test['l1'] = exog_test['l1'].iloc[10:30]
    exog_test['l2'] = exog_test['l2'].iloc[:40]

    exog_dict, exog_col_names = check_preprocess_exog_multiseries(
                                    input_series_is_dict = True,
                                    series_indexes       = series_indexes,
                                    series_col_names     = ['l1', 'l2', 'l3'],
                                    exog                 = exog_test,
                                    exog_dict            = {'l1': None, 'l2': None, 'l3': None}
                                )

    expected_exog_dict = {
        'l1': pd.DataFrame(
                  data    = exog_test['l1'].to_numpy(),
                  index   = pd.date_range(start='2000-01-11', periods=len(exog_test['l1']), freq='D'),
                  columns = ['exog_1', 'exog_2']
              ).astype({'exog_1': float}),
        'l2': pd.DataFrame(
                  data    = exog_test['l2'].to_numpy(),
                  index   = pd.date_range(start='2000-01-01', periods=len(exog_test['l2']), freq='D'),
                  columns = ['exog_1']
              ).astype({'exog_1': float}),
        'l3': None
    }
    expected_exog_col_names = ['exog_1', 'exog_2']
    
    assert isinstance(exog_dict, dict)
    assert list(exog_dict.keys()) == ['l1', 'l2', 'l3']
    for k in exog_dict:
        if k == 'l3':
            assert exog_dict[k] is None
        else:
            assert isinstance(exog_dict[k], pd.DataFrame)
            pd.testing.assert_frame_equal(exog_dict[k], expected_exog_dict[k])
            index_intersection = (
                series_dict[k].index.intersection(exog_dict[k].index)
            )
            assert len(index_intersection) == len(exog_dict[k])

    assert len(set(exog_col_names) - set(expected_exog_col_names)) == 0


# Unit test plot history method
# ==============================================================================
import pandas as pd
import numpy as np
import pytest
from skforecast.ForecasterRnn import ForecasterRnn
import tensorflow as tf
from skforecast.ForecasterRnn.utils import create_and_compile_model
import matplotlib.pyplot as plt

series = pd.DataFrame(
    {
        "1": pd.Series(np.arange(50)),
        "2": pd.Series(np.arange(50)),
        "3": pd.Series(np.arange(50)),
    }
)
lags = 3
steps = 4
levels = ["1", "2"]
activation = "relu"
optimizer = tf.keras.optimizers.Adam(learning_rate=0.01)
loss = tf.keras.losses.MeanSquaredError()
recurrent_units = 100
dense_units = [128, 64]


model = create_and_compile_model(
    series=series,
    lags=lags,
    steps=steps,
    levels=levels,
    recurrent_units=recurrent_units,
    dense_units=dense_units,
    activation=activation,
    optimizer=optimizer,
    loss=loss,
)
forecaster = ForecasterRnn(model, levels)
forecaster.fit(series)


def test_plot_history_with_val_loss():
    """
    Test case for the plot_history method
    """
    # Call the plot_history method
    fig, ax = plt.subplots()
    forecaster.plot_history(ax=ax)

    # Assert that the figure is of type matplotlib.figure.Figure
    assert isinstance(fig, plt.Figure)

    # Assert that the plot contains the training loss curve
    assert len(fig.axes[0].lines) == 1
    assert fig.axes[0].lines[0].get_label() == "Training Loss"

    # Assert that the plot contains the validation loss curve
    assert len(fig.axes[0].lines) == 1
    assert fig.axes[0].lines[0].get_label() == "Training Loss"



# Unit test __init__ ForecasterAutoregMultiVariate
# ==============================================================================
import re
import pytest
import numpy as np
import pandas as pd
from skforecast.ForecasterAutoregMultiVariate import ForecasterAutoregMultiVariate
from sklearn.linear_model import LinearRegression


def test_init_TypeError_when_level_is_not_a_str():
    """
    Test TypeError is raised when level is not a str.
    """
    level = 5
    err_msg = re.escape(f"`level` argument must be a str. Got {type(level)}.")
    with pytest.raises(TypeError, match = err_msg):
        ForecasterAutoregMultiVariate(LinearRegression(), level=level, lags=2, steps=3)


def test_init_TypeError_when_steps_is_not_int():
    """
    Test TypeError is raised when steps is not an int.
    """
    steps = 'not_valid_type'
    err_msg = re.escape(
                f"`steps` argument must be an int greater than or equal to 1. "
                f"Got {type(steps)}."
            )
    with pytest.raises(TypeError, match = err_msg):
        ForecasterAutoregMultiVariate(LinearRegression(), level='l1', lags=2, steps=steps)


def test_init_ValueError_when_steps_is_less_than_1():
    """
    Test ValueError is raised when steps is less than 1.
    """
    steps = 0
    err_msg = re.escape(f"`steps` argument must be greater than or equal to 1. Got {steps}.")
    with pytest.raises(ValueError, match = err_msg):
        ForecasterAutoregMultiVariate(LinearRegression(), level='l1', lags=2, steps=steps)


@pytest.mark.parametrize("lags, max_lag",
                         [(3, 3), 
                          ({'l1': 3, 'l2': 4}, 4), 
                          ({'l1': None, 'l2': 5}, 5)],
                         ids = lambda value : f'lags: {value}')
def test_init_max_lag_stored(lags, max_lag):
    """
    Test max_lag is equal to the maximum lag.
    """
    forecaster = ForecasterAutoregMultiVariate(LinearRegression(), level='l1', 
                                               lags=lags, steps=2)
    
    assert forecaster.max_lag == max_lag
    assert forecaster.window_size == max_lag


@pytest.mark.parametrize("n_jobs", 
                         [1.0, 'not_int_auto'], 
                         ids = lambda value : f'n_jobs: {value}')
def test_init_TypeError_when_n_jobs_not_int_or_auto(n_jobs):
    """
    Test TypeError is raised in when n_jobs is not an integer or 'auto'.
    """
    err_msg = re.escape(f"`n_jobs` must be an integer or `'auto'`. Got {type(n_jobs)}.")
    with pytest.raises(TypeError, match = err_msg):
        ForecasterAutoregMultiVariate(LinearRegression(), level='l1', lags=2, 
                                      steps=2, n_jobs=n_jobs)


# Unit test _create_lags ForecasterAutoreg
# ==============================================================================
import re
import pytest
import numpy as np
import pandas as pd
from skforecast.ForecasterAutoreg import ForecasterAutoreg
from sklearn.linear_model import LinearRegression
    
    
def test_create_lags_exception_when_len_of_y_is_lower_than_maximum_lag():
    """
    Test exception is raised when length of y is lower than maximum lag included
    in the forecaster.
    """
    y = pd.Series(np.arange(5), name='y')
    forecaster = ForecasterAutoreg(LinearRegression(), lags=10)
    err_msg = re.escape(
                (f'The maximum lag ({forecaster.max_lag}) must be less than the length '
                 f'of the series ({len(y)}).')
              )
    with pytest.raises(ValueError, match = err_msg):
        forecaster._create_lags(y=y)


def test_create_lags_output():
    """
    Test matrix of lags is created properly when lags=3 and y=np.arange(10).
    """
    forecaster = ForecasterAutoreg(LinearRegression(), lags=3)
    results = forecaster._create_lags(y=np.arange(10))
    expected = (np.array([[2., 1., 0.],
                          [3., 2., 1.],
                          [4., 3., 2.],
                          [5., 4., 3.],
                          [6., 5., 4.],
                          [7., 6., 5.],
                          [8., 7., 6.]]),
               np.array([3., 4., 5., 6., 7., 8., 9.]))

    assert (results[0] == expected[0]).all()
    assert (results[1] == expected[1]).all()


def test_create_lags_output_interspersed_lags():
    """
    Test matrix of lags is is a list with interspersed lags.
    """
    forecaster = ForecasterAutoreg(LinearRegression(), lags=[2, 3])
    results = forecaster._create_lags(y=np.arange(10))
    expected = (np.array([[1., 0.],
                          [2., 1.],
                          [3., 2.],
                          [4., 3.],
                          [5., 4.],
                          [6., 5.],
                          [7., 6.]]),
               np.array([3., 4., 5., 6., 7., 8., 9.]))

    assert (results[0] == expected[0]).all()
    assert (results[1] == expected[1]).all()


# Unit test predict_dist ForecasterAutoreg
# ==============================================================================
import numpy as np
import pandas as pd
from skforecast.ForecasterAutoreg import ForecasterAutoreg
from sklearn.linear_model import LinearRegression
from sklearn.preprocessing import StandardScaler
from scipy.stats import norm

# Fixtures
from .fixtures_ForecasterAutoreg import y
from .fixtures_ForecasterAutoreg import exog
from .fixtures_ForecasterAutoreg import exog_predict


def test_predict_dist_output_when_forecaster_is_LinearRegression_steps_is_2_in_sample_residuals_True_exog_and_transformer():
    """
    Test output of predict_dist when regressor is LinearRegression,
    2 steps are predicted, using in-sample residuals, exog is included and both
    inputs are transformed.
    """
    forecaster = ForecasterAutoreg(
                     regressor        = LinearRegression(),
                     lags             = 3,
                     transformer_y    = StandardScaler(),
                     transformer_exog = StandardScaler(),
                 )

    forecaster.fit(y=y, exog=exog)
    results = forecaster.predict_dist(
                  steps               = 2,
                  exog                = exog_predict,
                  distribution        = norm,
                  n_boot              = 4,
                  in_sample_residuals = True
              )
    
    expected = pd.DataFrame(
                   data    = np.array([[0.8123156 , 0.22809656],
                                       [0.26405621, 0.22123895]]),
                   columns = ['loc', 'scale'],
                   index   = pd.RangeIndex(start=50, stop=52)
               )
    
    pd.testing.assert_frame_equal(expected, results)


def test_predict_dist_output_when_forecaster_is_LinearRegression_steps_is_2_in_sample_residuals_False_exog_and_transformer():
    """
    Test output of predict_dist when regressor is LinearRegression,
    2 steps are predicted, using out-sample residuals, exog is included and both
    inputs are transformed.
    """
    forecaster = ForecasterAutoreg(
                     regressor        = LinearRegression(),
                     lags             = 3,
                     transformer_y    = StandardScaler(),
                     transformer_exog = StandardScaler(),
                 )
    
    forecaster.fit(y=y, exog=exog)
    forecaster.out_sample_residuals = forecaster.in_sample_residuals
    results = forecaster.predict_dist(
                  steps               = 2,
                  exog                = exog_predict,
                  distribution        = norm,
                  n_boot              = 4,
                  in_sample_residuals = False
              )
    
    expected = pd.DataFrame(
                   data    = np.array([[0.8123156 , 0.22809656],
                                       [0.26405621, 0.22123895]]),
                   columns = ['loc', 'scale'],
                   index   = pd.RangeIndex(start=50, stop=52)
               )

    pd.testing.assert_frame_equal(expected, results)


# Unit test set_fit_kwargs ForecasterAutoregMultiSeriesCustom
# ==============================================================================
from skforecast.ForecasterAutoregMultiSeriesCustom import ForecasterAutoregMultiSeriesCustom
from lightgbm import LGBMRegressor

def create_predictors(y): # pragma: no cover
    """
    Create first 5 lags of a time series.
    """

    lags = y[-1:-6:-1]

    return lags


def test_set_fit_kwargs():
    """
    Test set_fit_kwargs method.
    """
    forecaster = ForecasterAutoregMultiSeriesCustom(
                     regressor      = LGBMRegressor(),
                     fun_predictors = create_predictors,
                     window_size    = 5,
                     fit_kwargs     = {'categorical_feature': 'auto'}
                 )
    
    new_fit_kwargs = {'categorical_feature': ['exog']}
    forecaster.set_fit_kwargs(new_fit_kwargs)
    results = forecaster.fit_kwargs

    expected = {'categorical_feature': ['exog']}

    assert results == expected


# Unit test __init__ ForecasterAutoregCustom
# ==============================================================================
import re
import pytest
import pandas as pd
import numpy as np
from skforecast.ForecasterAutoregCustom import ForecasterAutoregCustom
from skforecast.exceptions import IgnoredArgumentWarning
from sklearn.linear_model import LinearRegression
from sklearn.neighbors import KNeighborsRegressor


def create_predictors(y): # pragma: no cover
    """
    Create first 5 lags of a time series.
    """
    
    lags = y[-1:-6:-1]
    
    return lags  


@pytest.mark.parametrize("window_size", 
                         [0, 0.5, 1.5, 'not_int'], 
                         ids = lambda ws : f'window_size: {ws}')
def test_init_ValueError_when_window_size_argument_is_not_int_or_greater_than_0(window_size):
    """
    Test ValueError is raised when window_size is not an int or greater than 0.
    """
    err_msg = re.escape(
        (f"Argument `window_size` must be an integer equal to or "
         f"greater than 1. Got {window_size}.")
    )
    with pytest.raises(ValueError, match = err_msg):
         ForecasterAutoregCustom(
             regressor      = LinearRegression(),
             fun_predictors = create_predictors,
             window_size    = window_size
         )


@pytest.mark.parametrize("dif", 
                         [0, 0.5, 1.5, 'not_int'], 
                         ids = lambda dif : f'differentiation: {dif}')
def test_init_ValueError_when_differentiation_argument_is_not_int_or_greater_than_0(dif):
    """
    Test ValueError is raised when differentiation is not an int or greater than 0.
    """
    err_msg = re.escape(
        (f"Argument `differentiation` must be an integer equal to or "
         f"greater than 1. Got {dif}.")
    )
    with pytest.raises(ValueError, match = err_msg):
         ForecasterAutoregCustom(
             regressor       = LinearRegression(),
             fun_predictors  = create_predictors,
             window_size     = 5,
             differentiation = dif
         )


@pytest.mark.parametrize("dif", 
                         [1, 2], 
                         ids = lambda dif : f'differentiation: {dif}')
def test_init_window_size_is_increased_when_differentiation(dif):
    """
    Test window_size is increased when including differentiation.
    """
    window_size = 5

    forecaster = ForecasterAutoregCustom(
                     regressor       = LinearRegression(),
                     fun_predictors  = create_predictors,
                     window_size     = window_size,
                     differentiation = dif
                 )
    
    assert forecaster.window_size == window_size
    assert forecaster.window_size_diff == window_size + dif


def test_init_TypeError_when_fun_predictors_argument_is_not_Callable():
    """
    Test TypeError is raised when fun_predictors is not a Callable.
    """
    fun_predictors = 'create_predictors'
    err_msg = re.escape(
        f"Argument `fun_predictors` must be a Callable. Got {type(fun_predictors)}."
    )
    with pytest.raises(TypeError, match = err_msg):
         ForecasterAutoregCustom(
             regressor      = LinearRegression(),
             fun_predictors = fun_predictors,
             window_size    = 5
         )


def test_init_TypeError_when_weight_func_argument_is_not_Callable():
    """
    Test TypeError is raised when weight_func is not a callable.
    """
    weight_func = '---'
    err_msg = re.escape(
        f"Argument `weight_func` must be a Callable. Got {type(weight_func)}."
    )
    with pytest.raises(TypeError, match = err_msg):
         ForecasterAutoregCustom(
             regressor      = LinearRegression(),
             fun_predictors = create_predictors,
             window_size    = 5,
             weight_func    = weight_func
         )


def test_init_IgnoredArgumentWarning_when_series_weights_is_provided_and_regressor_has_not_sample_weights():
    """
    Test IgnoredArgumentWarning is raised when weight_func is provided but the 
    regressor has no argument sample_weights in his fit method.
    """
    def weight_func(index): # pragma: no cover
         return np.ones_like(index)
 
    warn_msg = re.escape(
        ("Argument `weight_func` is ignored since regressor KNeighborsRegressor() "
         "does not accept `sample_weight` in its `fit` method.")
    )
    with pytest.warns(IgnoredArgumentWarning, match = warn_msg):
         ForecasterAutoregCustom(
             regressor      = KNeighborsRegressor(),
             fun_predictors = create_predictors,
             window_size    = 5,
             weight_func    = weight_func
         )


# Unit test _create_train_X_y_single_series ForecasterAutoregMultiSeriesCustom
# ==============================================================================
import re
import pytest
import numpy as np
import pandas as pd
from sklearn.base import clone
from sklearn.linear_model import LinearRegression
from sklearn.preprocessing import StandardScaler
from sklearn.preprocessing import MinMaxScaler
from skforecast.ForecasterAutoregMultiSeriesCustom import ForecasterAutoregMultiSeriesCustom

def create_predictors_3(y): # pragma: no cover
    """
    Create first 3 lags of a time series.
    """
    lags = y[-1:-4:-1]

    return lags

def create_predictors_5(y): # pragma: no cover
    """
    Create first 5 lags of a time series.
    """
    lags = y[-1:-6:-1]

    return lags


@pytest.mark.parametrize("differentiation", 
                         [None, 1, 2], 
                         ids = lambda dt : f'differentiation: {dt}')
def test_create_train_X_y_single_series_ValueError_when_forecaster_window_size_does_not_match_with_fun_predictors(
    differentiation,
):
    """
    Test ValueError is raised when the window needed by `fun_predictors()` does
    not correspond with the forecaster.window_size.
    """
    y = pd.Series(np.array([2, 56, 4.3, 23, 1, 8, 4.3, 104.1]), dtype=float, name="l1")

    forecaster = ForecasterAutoregMultiSeriesCustom(
        regressor=LinearRegression(),
        fun_predictors=create_predictors_3,
        window_size=2,
        differentiation=differentiation,
    )
    forecaster.transformer_series_ = {"l1": None}
    if differentiation is not None:
        forecaster.differentiator_ = {"l1": clone(forecaster.differentiator)}

    err_msg = re.escape(
        ("The `window_size` argument (2), declared when "
         "initializing the forecaster, does not correspond to the window "
         "used by `create_predictors_3`.")
    )
    with pytest.raises(ValueError, match=err_msg):
        forecaster._create_train_X_y_single_series(y=y, ignore_exog=True)


def test_create_train_X_y_single_series_ValueError_when_len_name_predictors_not_match_X_train_columns():
    """
    Test ValueError is raised when argument `name_predictors` has less values 
    than the number of columns of X_train.
    """
    y = pd.Series(np.arange(7, dtype=float), name='l1')

    forecaster = ForecasterAutoregMultiSeriesCustom(
                     regressor       = LinearRegression(),
                     fun_predictors  = create_predictors_3,
                     name_predictors = ['lag_1', 'lag_2'],
                     window_size     = 3
                 )
    forecaster.transformer_series_ = {'l1': None}

    err_msg = re.escape(
        ("The length of provided predictors names (`name_predictors`) do not "
         "match the number of columns created by `create_predictors_3`.")  
    )
    with pytest.raises(ValueError, match = err_msg):
        forecaster._create_train_X_y_single_series(y=y, ignore_exog=True)


@pytest.mark.parametrize("ignore_exog", 
                         [True, False], 
                         ids = lambda ie : f'ignore_exog: {ie}')
def test_create_train_X_y_single_series_output_when_series_and_exog_is_None(ignore_exog):
    """
    Test the output of _create_train_X_y_single_series when exog is None. 
    Check cases `ignore_exog` True and False.
    """
    y = pd.Series(np.arange(7, dtype=float), name='l1')
    
    forecaster = ForecasterAutoregMultiSeriesCustom(
                     regressor      = LinearRegression(),
                     fun_predictors = create_predictors_3,
                     window_size    = 3
                 )
    forecaster.transformer_series_ = {'l1': StandardScaler()}
    results = forecaster._create_train_X_y_single_series(
                  y           = y,
                  ignore_exog = ignore_exog,
                  exog        = None
              )
    
    expected = (
        pd.DataFrame(
            data = np.array([[-0.5, -1. , -1.5, 'l1'],
                             [ 0. , -0.5, -1. , 'l1'],
                             [ 0.5,  0. , -0.5, 'l1'],
                             [ 1. ,  0.5,  0. , 'l1']]),
            index   = pd.RangeIndex(start=3, stop=7, step=1),
            columns = ['custom_predictor_0', 'custom_predictor_1', 'custom_predictor_2',
                       '_level_skforecast']
        ).astype({'custom_predictor_0': float, 'custom_predictor_1': float, 'custom_predictor_2': float}),
        pd.DataFrame(
            data    = np.nan,
            columns = ['_dummy_exog_col_to_keep_shape'],
            index   = pd.RangeIndex(start=3, stop=7, step=1)
        ),
        pd.Series(
            data  = np.array([0., 0.5, 1., 1.5]),
            index = pd.RangeIndex(start=3, stop=7, step=1),
            name  = 'y',
            dtype = float
        )
    )

    pd.testing.assert_frame_equal(results[0], expected[0])
    if ignore_exog:
        assert isinstance(results[1], type(None))
    else:
        pd.testing.assert_frame_equal(results[1], expected[1])
    pd.testing.assert_series_equal(results[2], expected[2])


def test_create_train_X_y_single_series_output_when_series_and_exog():
    """
    Test the output of _create_train_X_y_single_series when exog is a 
    pandas DataFrame with one column.
    """
    y = pd.Series(np.arange(10, dtype=float), name='l1')
    exog = pd.DataFrame(np.arange(100, 110, dtype=float), columns=['exog'])
    
    forecaster = ForecasterAutoregMultiSeriesCustom(
                     regressor          = LinearRegression(),
                     fun_predictors     = create_predictors_5,
                     window_size        = 5,
                     name_predictors    = ['lag_1', 'lag_2', 'lag_3', 'lag_4', 'lag_5'],
                     transformer_series = None
                 )
    forecaster.transformer_series_ = {'l1': None}
    results = forecaster._create_train_X_y_single_series(
                  y           = y,
                  ignore_exog = False,
                  exog        = exog
              )

    expected = (
        pd.DataFrame(
            data = np.array([[4., 3., 2., 1., 0., 'l1'],
                             [5., 4., 3., 2., 1., 'l1'],
                             [6., 5., 4., 3., 2., 'l1'],
                             [7., 6., 5., 4., 3., 'l1'],
                             [8., 7., 6., 5., 4., 'l1']]),
            index   = pd.RangeIndex(start=5, stop=10, step=1),
            columns = ['lag_1', 'lag_2', 'lag_3', 'lag_4', 'lag_5', 
                       '_level_skforecast']
        ).astype({'lag_1': float, 'lag_2': float, 'lag_3': float,
                  'lag_4': float, 'lag_5': float}),
        pd.DataFrame(
            data  = np.array([105., 106., 107., 108., 109.], dtype=float),
            index = pd.RangeIndex(start=5, stop=10, step=1),
            columns = ['exog']
        ),
        pd.Series(
            data  = np.array([5., 6., 7., 8., 9.], dtype=float),
            index = pd.RangeIndex(start=5, stop=10, step=1),
            name  = 'y'
        )
    )

    pd.testing.assert_frame_equal(results[0], expected[0])
    pd.testing.assert_frame_equal(results[1], expected[1])
    pd.testing.assert_series_equal(results[2], expected[2])


def test_create_train_X_y_single_series_output_when_series_10_and_exog_is_dataframe_of_category():
    """
    Test the output of _create_train_X_y_single_series when exog is a 
    pandas DataFrame with one column float and one column of cateogry.
    """
    y = pd.Series(np.arange(10, dtype=float), name='l1')
    exog = pd.DataFrame({'exog_1': np.arange(10, 20, dtype=float),
                         'exog_2': pd.Categorical(range(100, 110))})
    
    forecaster = ForecasterAutoregMultiSeriesCustom(
                     regressor          = LinearRegression(),
                     fun_predictors     = create_predictors_5,
                     window_size        = 5,
                     name_predictors    = ['lag_1', 'lag_2', 'lag_3', 'lag_4', 'lag_5'],
                     transformer_series = None
                 )
    forecaster.transformer_series_ = {'l1': None}
    results = forecaster._create_train_X_y_single_series(
                  y           = y,
                  ignore_exog = False,
                  exog        = exog
              )

    expected = (
        pd.DataFrame(
            data = np.array([[4., 3., 2., 1., 0., 'l1'],
                             [5., 4., 3., 2., 1., 'l1'],
                             [6., 5., 4., 3., 2., 'l1'],
                             [7., 6., 5., 4., 3., 'l1'],
                             [8., 7., 6., 5., 4., 'l1']]),
            index   = pd.RangeIndex(start=5, stop=10, step=1),
            columns = ['lag_1', 'lag_2', 'lag_3', 'lag_4', 'lag_5', 
                       '_level_skforecast']
        ).astype({'lag_1': float, 'lag_2': float, 'lag_3': float,
                  'lag_4': float, 'lag_5': float}),
        pd.DataFrame(
            data  = np.array([15., 16., 17., 18., 19.], dtype=float),
            index = pd.RangeIndex(start=5, stop=10, step=1),
            columns = ['exog_1']
        ).assign(exog_2 = pd.Categorical([105, 106, 107, 108, 109], categories=range(100, 110))),
        pd.Series(
            data  = np.array([5., 6., 7., 8., 9.], dtype=float),
            index = pd.RangeIndex(start=5, stop=10, step=1),
            name  = 'y'
        )
    )

    pd.testing.assert_frame_equal(results[0], expected[0])
    pd.testing.assert_frame_equal(results[1], expected[1])
    pd.testing.assert_series_equal(results[2], expected[2])


def test_create_train_X_y_single_series_output_when_series_and_exog_is_DataFrame_datetime_index():
    """
    Test the output of _create_train_X_y_single_series when series and 
    exog is a pandas dataframe with two columns and datetime index.
    """
    y = pd.Series(np.arange(7, dtype=float), name='l1')
    y.index = pd.date_range("1990-01-01", periods=7, freq='D')
    exog = pd.DataFrame({
               'exog_1' : np.arange(100, 107, dtype=float),
               'exog_2' : np.arange(1000, 1007, dtype=float)},
               index = pd.date_range("1990-01-01", periods=7, freq='D')
           )
    
    forecaster = ForecasterAutoregMultiSeriesCustom(
                     regressor          = LinearRegression(),
                     fun_predictors     = create_predictors_3,
                     window_size        = 3,
                     transformer_series = None
                 )
    forecaster.transformer_series_ = {'l1': None}
    results = forecaster._create_train_X_y_single_series(
                  y           = y,
                  ignore_exog = False,
                  exog        = exog
              )
    
    expected = (
        pd.DataFrame(
            data = np.array([[2.0, 1.0, 0.0, 'l1'],
                             [3.0, 2.0, 1.0, 'l1'],
                             [4.0, 3.0, 2.0, 'l1'],
                             [5.0, 4.0, 3.0, 'l1']]),
            index   = pd.date_range("1990-01-04", periods=4, freq='D'),
            columns = ['custom_predictor_0', 'custom_predictor_1', 'custom_predictor_2', '_level_skforecast']
        ).astype({'custom_predictor_0': float, 'custom_predictor_1': float, 'custom_predictor_2': float}),
        pd.DataFrame(
            data = np.array([[103., 1003.],
                             [104., 1004.],
                             [105., 1005.],
                             [106., 1006.]]),
            index   = pd.date_range("1990-01-04", periods=4, freq='D'),
            columns = ['exog_1', 'exog_2']
        ),
        pd.Series(
            data  = np.array([3., 4., 5., 6.], dtype=float),
            index = pd.date_range("1990-01-04", periods=4, freq='D'),
            name  = 'y'
        )
    )

    pd.testing.assert_frame_equal(results[0], expected[0])
    pd.testing.assert_frame_equal(results[1], expected[1])
    pd.testing.assert_series_equal(results[2], expected[2])


def test_create_train_X_y_single_series_output_when_series_and_exog_is_DataFrame_with_NaNs():
    """
    Test the output of _create_train_X_y_single_series when series and 
    exog is a pandas dataframe with two columns and NaNs in between.
    """
    y = pd.Series(np.arange(10, dtype=float), name='l1')
    y.iloc[6:8] = np.nan
    exog = pd.DataFrame({
               'exog_1' : np.arange(100, 110, dtype=float),
               'exog_2' : np.arange(1000, 1010, dtype=float)}
           )
    exog.iloc[2:7, 0] = np.nan
    
    forecaster = ForecasterAutoregMultiSeriesCustom(
                     regressor          = LinearRegression(),
                     fun_predictors     = create_predictors_5,
                     window_size        = 5,
                     name_predictors    = ['lag_1', 'lag_2', 'lag_3', 'lag_4', 'lag_5'],
                     transformer_series = None
                 )
    forecaster.transformer_series_ = {'l1': None}
    results = forecaster._create_train_X_y_single_series(
                  y           = y,
                  ignore_exog = False,
                  exog        = exog
              )

    expected = (
        pd.DataFrame(
            data = np.array([[4., 3., 2., 1., 0., 'l1'],
                             [5., 4., 3., 2., 1., 'l1'],
                             [np.nan, 5., 4., 3., 2., 'l1'],
                             [np.nan, np.nan, 5., 4., 3., 'l1'],
                             [8., np.nan, np.nan, 5., 4., 'l1']]),
            index   = pd.RangeIndex(start=5, stop=10, step=1),
            columns = ['lag_1', 'lag_2', 'lag_3', 'lag_4', 'lag_5', 
                       '_level_skforecast']
        ).astype({'lag_1': float, 'lag_2': float, 'lag_3': float,
                  'lag_4': float, 'lag_5': float}),
        pd.DataFrame(
            data = np.array([[np.nan, 1005.],
                             [np.nan, 1006.],
                             [107., 1007.],
                             [108., 1008.],
                             [109., 1009.]]),
            index = pd.RangeIndex(start=5, stop=10, step=1),
            columns = ['exog_1', 'exog_2']
        ),
        pd.Series(
            data  = np.array([5., np.nan, np.nan, 8., 9.], dtype=float),
            index = pd.RangeIndex(start=5, stop=10, step=1),
            name  = 'y'
        )
    )

    pd.testing.assert_frame_equal(results[0], expected[0])
    pd.testing.assert_frame_equal(results[1], expected[1])
    pd.testing.assert_series_equal(results[2], expected[2])


def test_create_train_X_y_single_series_output_when_transformer_and_fitted():
    """
    Test the output of _create_train_X_y_single_series when Forecaster as
    already been fitted, transformer is MinMaxScaler() and has a different 
    series as input.
    """
    y = pd.Series(np.arange(9, dtype=float), name='l1')
    transformer = MinMaxScaler()
    transformer.fit(y.values.reshape(-1, 1))
    
    forecaster = ForecasterAutoregMultiSeriesCustom(
                     regressor          = LinearRegression(),
                     fun_predictors     = create_predictors_3,
                     window_size        = 3,
                     transformer_series = None
                 )
    forecaster.transformer_series_ = {'l1': transformer}
    forecaster.fitted = True

    new_y = pd.Series(np.arange(10, 19, dtype=float), name='l1')
    results = forecaster._create_train_X_y_single_series(y=new_y, ignore_exog=True)
    
    expected = (
        pd.DataFrame(
            data = np.array([[1.5  , 1.375, 1.25 , 'l1'],
                             [1.625, 1.5  , 1.375, 'l1'],
                             [1.75 , 1.625, 1.5  , 'l1'],
                             [1.875, 1.75 , 1.625, 'l1'],
                             [2.   , 1.875, 1.75 , 'l1'],
                             [2.125, 2.   , 1.875, 'l1']]),
            index   = pd.RangeIndex(start=3, stop=9, step=1),
            columns = ['custom_predictor_0', 'custom_predictor_1', 'custom_predictor_2', '_level_skforecast']
        ).astype({'custom_predictor_0': float, 'custom_predictor_1': float, 'custom_predictor_2': float}),
        None,
        pd.Series(
            data  = np.array([1.625, 1.75, 1.875, 2., 2.125, 2.25]),
            index = pd.RangeIndex(start=3, stop=9, step=1),
            name  = 'y',
            dtype = float
        )
    )

    pd.testing.assert_frame_equal(results[0], expected[0])
    assert isinstance(results[1], type(None))
    pd.testing.assert_series_equal(results[2], expected[2])


@pytest.mark.parametrize("fitted", 
                         [True, False], 
                         ids = lambda fitted : f'fitted: {fitted}')
def test_create_train_X_y_single_series_output_when_series_and_exog_and_differentitation_1(fitted):
    """
    Test the output of _create_train_X_y_single_series when exog is a 
    pandas DataFrame with one column and differentiation=1.
    """
    y = pd.Series(np.arange(10, dtype=float), name='l1')
    exog = pd.DataFrame(np.arange(100, 110, dtype=float), columns=['exog'])
    
    forecaster = ForecasterAutoregMultiSeriesCustom(
                     regressor          = LinearRegression(),
                     fun_predictors     = create_predictors_5,
                     window_size        = 5,
                     name_predictors    = ['lag_1', 'lag_2', 'lag_3', 'lag_4', 'lag_5'],
                     transformer_series = None,
                     differentiation    = 1
                 )
    forecaster.transformer_series_ = {'l1': None}
    forecaster.differentiator_ = {'l1': clone(forecaster.differentiator)}
    forecaster.fitted = fitted

    results = forecaster._create_train_X_y_single_series(
                  y           = y,
                  ignore_exog = False,
                  exog        = exog
              )

    expected = (
        pd.DataFrame(
            data = np.array([[1., 1., 1., 1., 1., 'l1'],
                             [1., 1., 1., 1., 1., 'l1'],
                             [1., 1., 1., 1., 1., 'l1'],
                             [1., 1., 1., 1., 1., 'l1']]),
            index   = pd.RangeIndex(start=6, stop=10, step=1),
            columns = ['lag_1', 'lag_2', 'lag_3', 'lag_4', 'lag_5', 
                       '_level_skforecast']
        ).astype({'lag_1': float, 'lag_2': float, 'lag_3': float,
                  'lag_4': float, 'lag_5': float}),
        pd.DataFrame(
            data  = np.array([106., 107., 108., 109.], dtype=float),
            index = pd.RangeIndex(start=6, stop=10, step=1),
            columns = ['exog']
        ),
        pd.Series(
            data  = np.array([1., 1., 1., 1.], dtype=float),
            index = pd.RangeIndex(start=6, stop=10, step=1),
            name  = 'y'
        )
    )

    pd.testing.assert_frame_equal(results[0], expected[0])
    pd.testing.assert_frame_equal(results[1], expected[1])
    pd.testing.assert_series_equal(results[2], expected[2])


@pytest.mark.parametrize("fitted", 
                         [True, False], 
                         ids = lambda fitted : f'fitted: {fitted}')
def test_create_train_X_y_single_series_output_when_series_and_exog_and_differentitation_2(fitted):
    """
    Test the output of _create_train_X_y_single_series when exog is a 
    pandas DataFrame with one column and differentiation=2.
    """
    y = pd.Series(np.arange(10, dtype=float), name='l1')
    exog = pd.DataFrame(np.arange(100, 110, dtype=float), columns=['exog'])
    
    forecaster = ForecasterAutoregMultiSeriesCustom(
                     regressor          = LinearRegression(),
                     fun_predictors     = create_predictors_5,
                     window_size        = 5,
                     name_predictors    = ['lag_1', 'lag_2', 'lag_3', 'lag_4', 'lag_5'],
                     transformer_series = None,
                     differentiation    = 2
                 )
    forecaster.transformer_series_ = {'l1': None}
    forecaster.differentiator_ = {'l1': clone(forecaster.differentiator)}
    forecaster.fitted = fitted

    results = forecaster._create_train_X_y_single_series(
                  y           = y,
                  ignore_exog = False,
                  exog        = exog
              )

    expected = (
        pd.DataFrame(
            data = np.array([[0., 0., 0., 0., 0., 'l1'],
                             [0., 0., 0., 0., 0., 'l1'],
                             [0., 0., 0., 0., 0., 'l1']]),
            index   = pd.RangeIndex(start=7, stop=10, step=1),
            columns = ['lag_1', 'lag_2', 'lag_3', 'lag_4', 'lag_5', 
                       '_level_skforecast']
        ).astype({'lag_1': float, 'lag_2': float, 'lag_3': float,
                  'lag_4': float, 'lag_5': float}),
        pd.DataFrame(
            data  = np.array([107., 108., 109.], dtype=float),
            index = pd.RangeIndex(start=7, stop=10, step=1),
            columns = ['exog']
        ),
        pd.Series(
            data  = np.array([0., 0., 0.], dtype=float),
            index = pd.RangeIndex(start=7, stop=10, step=1),
            name  = 'y'
        )
    )

    pd.testing.assert_frame_equal(results[0], expected[0])
    pd.testing.assert_frame_equal(results[1], expected[1])
    pd.testing.assert_series_equal(results[2], expected[2])



# Unit test fit Sarimax
# ==============================================================================
import re
import pytest
import numpy as np
import pandas as pd
from sklearn.exceptions import NotFittedError
from skforecast.Sarimax import Sarimax
from statsmodels.tsa.statespace.sarimax import SARIMAX
from statsmodels.tsa.statespace.sarimax import SARIMAXResults
from statsmodels.tsa.statespace.sarimax import SARIMAXResultsWrapper

# Fixtures
from .fixtures_Sarimax import y
from .fixtures_Sarimax import y_lw
from .fixtures_Sarimax import exog
from .fixtures_Sarimax import exog_lw
from .fixtures_Sarimax import exog_predict
from .fixtures_Sarimax import exog_lw_predict
from .fixtures_Sarimax import y_datetime
from .fixtures_Sarimax import exog_datetime
from .fixtures_Sarimax import exog_predict_datetime

# Fixtures numpy
y_numpy = y.to_numpy()
exog_numpy = exog.to_numpy()
exog_predict_numpy = exog_predict.to_numpy()

y_lw_numpy = y_lw.to_numpy()
exog_lw_numpy = exog_lw.to_numpy()
exog_lw_predict_numpy = exog_lw_predict.to_numpy()


def test_decorator_check_fitted():
    """
    Test Sarimax decorator check if regressor is fitted.
    """

    sarimax = Sarimax(order=(1, 1, 1))

    err_msg = re.escape(
        ("Sarimax instance is not fitted yet. Call `fit` with "
         "appropriate arguments before using this method.")
    )
    with pytest.raises(NotFittedError, match = err_msg): 
        sarimax.predict(steps=5)
    with pytest.raises(NotFittedError, match = err_msg): 
        sarimax.append(y=y)
    with pytest.raises(NotFittedError, match = err_msg): 
        sarimax.apply(y=y)
    with pytest.raises(NotFittedError, match = err_msg): 
        sarimax.extend(y=y)
    with pytest.raises(NotFittedError, match = err_msg): 
        sarimax.params()
    with pytest.raises(NotFittedError, match = err_msg): 
        sarimax.summary()


def test_Sarimax_sarimax_params_are_stored_during_initialization():
    """
    Test if `sarimax._sarimax_params` are stored correctly during initialization.
    """

    sarimax = Sarimax()
    results = sarimax._sarimax_params

    expected_params = {
        'order': (1, 0, 0),
        'seasonal_order': (0, 0, 0, 0),
        'trend': None,
        'measurement_error': False,
        'time_varying_regression': False,
        'mle_regression': True,
        'simple_differencing': False,
        'enforce_stationarity': True,
        'enforce_invertibility': True,
        'hamilton_representation': False,
        'concentrate_scale': False,
        'trend_offset': 1,
        'use_exact_diffuse': False,
        'dates': None,
        'freq': None,
        'missing': 'none',
        'validate_specification': True,
        'method': 'lbfgs',
        'maxiter': 50,
        'start_params': None,
        'disp': False,
        'sm_init_kwargs': {},
        'sm_fit_kwargs': {},
        'sm_predict_kwargs': {}
    }

    assert isinstance(results, dict)
    assert results == expected_params


def test_Sarimax__repr__(capfd):
    """
    Check information printed by a Sarimax object.
    """

    sarimax = Sarimax(order=(1, 2, 3), seasonal_order=(4, 5, 6, 12))
    expected_out = "Sarimax(1,2,3)(4,5,6)[12]\n"

    print(sarimax)
    out, _ = capfd.readouterr()

    assert out == expected_out


@pytest.mark.parametrize("kwargs, _init_kwargs", 
                         [({'order': (1,1,1), 
                            'sm_init_kwargs': {'order': (2,2,2)}}, 
                           {'order': (1,1,1)}), 
                          ({'order': (2,2,2), 
                            'sm_init_kwargs': {'test': 'fake_param'}}, 
                           {'order': (2,2,2), 
                            'test': 'fake_param'})], 
                         ids = lambda values : f'kwargs, _init_kwargs: {values}')
def test_Sarimax_consolidate_kwargs_init_kwargs(kwargs, _init_kwargs):
    """
    Test if `sarimax._init_kwargs` are correctly consolidate.
    """

    sarimax = Sarimax(**kwargs)
    results = (sarimax._init_kwargs, sarimax.sm_init_kwargs)

    # First expected _init_kwargs, second sm_init_kwargs
    expected = ({
        'order': (1, 0, 0),
        'seasonal_order': (0, 0, 0, 0),
        'trend': None,
        'measurement_error': False,
        'time_varying_regression': False,
        'mle_regression': True,
        'simple_differencing': False,
        'enforce_stationarity': True,
        'enforce_invertibility': True,
        'hamilton_representation': False,
        'concentrate_scale': False,
        'trend_offset': 1,
        'use_exact_diffuse': False,
        'dates': None,
        'freq': None,
        'missing': 'none',
        'validate_specification': True
    },
    kwargs['sm_init_kwargs']
    )
    expected[0].update(_init_kwargs)

    assert all(isinstance(x, dict) for x in results)
    assert results[0] == expected[0]
    assert results[1] == expected[1]


@pytest.mark.parametrize("kwargs, _fit_kwargs", 
                         [({'maxiter': 100, 
                            'sm_fit_kwargs': {'maxiter': 200}}, 
                           {'maxiter': 100}), 
                          ({'maxiter': 200, 
                            'sm_fit_kwargs': {'test': 'fake_param'}}, 
                           {'maxiter': 200, 
                            'test': 'fake_param'})], 
                         ids = lambda values : f'kwargs, _fit_kwargs: {values}')
def test_Sarimax_consolidate_kwargs_fit_kwargs(kwargs, _fit_kwargs):
    """
    Test if `sarimax._fit_kwargs` are correctly consolidate.
    """

    sarimax = Sarimax(**kwargs)
    results = (sarimax._fit_kwargs, sarimax.sm_fit_kwargs)

    # First expected _fit_kwargs, second sm_fit_kwargs
    expected = ({
        'method': 'lbfgs',
        'maxiter': 50,
        'start_params': None,
        'disp': False
    },
    kwargs['sm_fit_kwargs']
    )
    expected[0].update(_fit_kwargs)

    assert all(isinstance(x, dict) for x in results)
    assert results[0] == expected[0]
    assert results[1] == expected[1]


@pytest.mark.parametrize("kwargs, _predict_kwargs", 
                         [({'sm_predict_kwargs': {'test': 'fake_param'}}, 
                           {'test': 'fake_param'})], 
                         ids = lambda values : f'kwargs, _predict_kwargs: {values}')
def test_Sarimax_consolidate_kwargs_predict_kwargs(kwargs, _predict_kwargs):
    """
    Test if `sarimax._predict_kwargs` are correctly consolidate.
    """

    sarimax = Sarimax(**kwargs)
    results = (sarimax._predict_kwargs, sarimax.sm_predict_kwargs)

    # First expected _predict_kwargs, second sm_predict_kwargs
    expected = (
    {},
    kwargs['sm_predict_kwargs']
    )
    expected[0].update(_predict_kwargs)

    assert all(isinstance(x, dict) for x in results)
    assert results[0] == expected[0]
    assert results[1] == expected[1]


def test_Sarimax_create_sarimax():
    """
    Test statsmodels SARIMAX is correctly create with _create_sarimax.
    """

    sarimax = Sarimax(order=(1, 1, 1))
    sarimax._create_sarimax(endog=y)
    sarimax_stats = sarimax.sarimax

    assert isinstance(sarimax_stats, SARIMAX)
    assert sarimax_stats.order == (1, 1, 1)


def test_Sarimax_fit_with_numpy():
    """
    Test Sarimax fit with numpy `y` and `exog`.
    """

    sarimax = Sarimax(order=(1, 1, 1))

    assert sarimax.output_type is None
    assert sarimax.sarimax_res is None
    assert sarimax.fitted == False
    assert sarimax.training_index is None

    sarimax.fit(y=y_numpy, exog=exog_numpy)

    assert sarimax.output_type == 'numpy'
    assert isinstance(sarimax.sarimax_res, SARIMAXResultsWrapper)
    assert sarimax.fitted == True
    assert sarimax.training_index == None


@pytest.mark.parametrize("y, exog", 
                         [(y, exog),
                          (y_datetime, exog_datetime),
                          (y.to_frame(), exog.to_frame()),
                          (y_datetime.to_frame(), exog_datetime.to_frame())], 
                         ids = lambda values : f'y, exog: {type(values)}')
def test_Sarimax_fit_with_pandas(y, exog):
    """
    Test Sarimax fit with pandas `y` and `exog`.
    """

    sarimax = Sarimax(order=(1, 1, 1))

    assert sarimax.output_type is None
    assert sarimax.sarimax_res is None
    assert sarimax.fitted == False
    assert sarimax.training_index is None

    sarimax.fit(y=y, exog=exog)

    assert sarimax.output_type == 'pandas'
    assert isinstance(sarimax.sarimax_res, SARIMAXResultsWrapper)
    assert sarimax.fitted == True
    pd.testing.assert_index_equal(sarimax.training_index, y.index)


def test_Sarimax_predict_with_numpy():
    """
    Test Sarimax predict with numpy `y` and `exog`. It is tested with the 
    skforecast wrapper and the statsmodels SARIMAX.
    """

    # skforecast
    sarimax = Sarimax(order=(1, 1, 1), maxiter=1000, method='cg', disp=False)
    sarimax.fit(y=y_numpy, exog=exog_numpy)
    preds_skforecast = sarimax.predict(steps=10, exog=exog_predict_numpy)

    # statsmodels
    sarimax = SARIMAX(order=(1, 1, 1), endog=y_numpy, exog=exog_numpy)               
    sarimax_res = sarimax.fit(maxiter=1000, method='cg', disp=False)
    preds_statsmodels = sarimax_res.get_forecast(
                            steps = 10, 
                            exog  = exog_predict_numpy
                        ).predicted_mean

    # expected = np.array([0.54251563, 0.57444558, 0.60419802, 0.63585684, 0.67238625,
    #                      0.70992311, 0.76089838, 0.7855481 , 0.73117064, 0.669226  ])

    # np.testing.assert_almost_equal(preds_skforecast, expected)
    # np.testing.assert_almost_equal(preds_statsmodels, expected)
    np.testing.assert_almost_equal(preds_skforecast, preds_statsmodels)


@pytest.mark.parametrize("y, exog, exog_predict", 
                         [(y, exog, exog_predict),
                          (y_datetime, exog_datetime, 
                           exog_predict_datetime),
                          (y.to_frame(), exog.to_frame(), 
                           exog_predict.to_frame()),
                          (y_datetime.to_frame(), exog_datetime.to_frame(), 
                           exog_predict_datetime.to_frame())], 
                         ids = lambda values : f'y, exog, exog_predict: {type(values)}')
def test_Sarimax_predict_with_pandas(y, exog, exog_predict):
    """
    Test Sarimax predict with pandas `y` and `exog`. It is tested with the 
    skforecast wrapper and the statsmodels SARIMAX.
    """

    # skforecast
    sarimax = Sarimax(order=(1, 1, 1), maxiter=1000, method='cg', disp=False)
    sarimax.fit(y=y, exog=exog)
    preds_skforecast = sarimax.predict(steps=10, exog=exog_predict)

    # statsmodels
    sarimax = SARIMAX(order=(1, 1, 1), endog=y, exog=exog)               
    sarimax_res = sarimax.fit(maxiter=1000, method='cg', disp=False)
    preds_statsmodels = sarimax_res.get_forecast(
                            steps = 10, 
                            exog  = exog_predict
                        ).predicted_mean.rename("pred").to_frame()

    # if isinstance(y.index, pd.RangeIndex):
    #     idx = pd.RangeIndex(start=50, stop=60, step=1)
    # else:
    #     idx = pd.date_range(start='2050', periods=10, freq='A')
    #
    # expected = pd.DataFrame(
    #     data    = np.array([0.54251563, 0.57444558, 0.60419802, 0.63585684, 0.67238625,
    #                         0.70992311, 0.76089838, 0.7855481 , 0.73117064, 0.669226  ]),
    #     index   = idx,
    #     columns = ['pred']
    # )

    # pd.testing.assert_frame_equal(preds_skforecast, expected)
    # pd.testing.assert_frame_equal(preds_statsmodels, expected)
    pd.testing.assert_frame_equal(preds_skforecast, preds_statsmodels)


@pytest.mark.parametrize("y, exog, exog_predict", 
                         [(y_numpy, exog_numpy, exog_predict_numpy),
                          (y_datetime, exog_datetime, exog_predict_datetime),
                          (y_datetime.to_frame(), exog_datetime.to_frame(), 
                           exog_predict_datetime.to_frame())], 
                         ids = lambda values : f'y, exog, exog_predict: {type(values)}')
def test_Sarimax_predict_UserWarning_exog_length_greater_that_steps(y, exog, exog_predict):
    """
    Test Sarimax predict UserWarning is raised with numpy and pandas when length 
    exog is greater than the number of steps.
    """

    # skforecast
    sarimax = Sarimax(order=(1, 1, 1), maxiter=1000, method='cg', disp=False)
    sarimax.fit(y=y, exog=exog)

    steps = 5

    warn_msg = re.escape(
        ("When predicting using exogenous variables, the `exog` parameter "
         "must have the same length as the number of predicted steps. Since "
         "len(exog) > steps, only the first 5 observations are used.")
    )
    with pytest.warns(UserWarning, match = warn_msg):
        sarimax.predict(steps=steps, exog=exog_predict)


def test_Sarimax_predict_interval_with_numpy():
    """
    Test Sarimax predict with confidence intervals using numpy `y` and `exog`.  
    It is tested with the skforecast wrapper and the statsmodels SARIMAX.
    """

    alpha = 0.05

    # skforecast
    sarimax = Sarimax(order=(1, 1, 1), maxiter=1000, method='cg', disp=False)
    sarimax.fit(y=y_numpy, exog=exog_numpy)
    preds_skforecast = sarimax.predict(steps=10, exog=exog_predict_numpy,
                                       return_conf_int=True, alpha=alpha)

    # statsmodels
    sarimax = SARIMAX(order=(1, 1, 1), endog=y_numpy, exog=exog_numpy)               
    sarimax_res = sarimax.fit(maxiter=1000, method='cg', disp=False)
    preds_statsmodels = sarimax_res.get_forecast(
                            steps = 10, 
                            exog  = exog_predict_numpy
                        )
    preds_statsmodels = np.column_stack(
                            [preds_statsmodels.predicted_mean,
                             preds_statsmodels.conf_int(alpha=alpha)]
                        )

    # expected = np.array([[ 0.54251563, -1.19541248,  2.28044374],
    #                      [ 0.57444558, -1.81113178,  2.96002293],
    #                      [ 0.60419802, -2.26035953,  3.46875557],
    #                      [ 0.63585684, -2.62746432,  3.899178  ],
    #                      [ 0.67238625, -2.94147506,  4.28624756],
    #                      [ 0.70992311, -3.22142148,  4.64126771],
    #                      [ 0.76089838, -3.46329933,  4.98509609],
    #                      [ 0.7855481 , -3.7121082 ,  5.28320441],
    #                      [ 0.73117064, -4.02408271,  5.48642398],
    #                      [ 0.669226  , -4.3302976 ,  5.66874961]])

    # np.testing.assert_almost_equal(preds_skforecast, expected)
    # np.testing.assert_almost_equal(preds_statsmodels, expected)
    np.testing.assert_almost_equal(preds_skforecast, preds_statsmodels)


@pytest.mark.parametrize("y, exog, exog_predict", 
                         [(y, exog, exog_predict),
                          (y_datetime, exog_datetime, 
                           exog_predict_datetime),
                          (y.to_frame(), exog.to_frame(), 
                           exog_predict.to_frame()),
                          (y_datetime.to_frame(), exog_datetime.to_frame(), 
                           exog_predict_datetime.to_frame())], 
                         ids = lambda values : f'y, exog, exog_predict: {type(values)}')
def test_Sarimax_predict_interval_with_pandas(y, exog, exog_predict):
    """
    Test Sarimax predict with confidence intervals using pandas `y` and `exog`.  
    It is tested with the skforecast wrapper and the statsmodels SARIMAX.
    """

    alpha = 0.05

    # skforecast
    sarimax = Sarimax(order=(1, 1, 1), maxiter=1000, method='cg', disp=False)
    sarimax.fit(y=y, exog=exog)
    preds_skforecast = sarimax.predict(steps=10, exog=exog_predict,
                                       return_conf_int=True, alpha=alpha)

    # statsmodels
    sarimax = SARIMAX(order=(1, 1, 1), endog=y, exog=exog)               
    sarimax_res = sarimax.fit(maxiter=1000, method='cg', disp=False)
    preds_statsmodels = sarimax_res.get_forecast(
                            steps = 10, 
                            exog  = exog_predict
                        )
    preds_statsmodels = pd.concat((
                            preds_statsmodels.predicted_mean,
                            preds_statsmodels.conf_int(alpha=alpha)),
                            axis = 1
                        )
    preds_statsmodels.columns = ['pred', 'lower_bound', 'upper_bound']

    # if isinstance(y.index, pd.RangeIndex):
    #     idx = pd.RangeIndex(start=50, stop=60, step=1)
    # else:
    #     idx = pd.date_range(start='2050', periods=10, freq='A')
    
    # expected = pd.DataFrame(
    #     data    = np.array([[ 0.54251563, -1.19541248,  2.28044374],
    #                         [ 0.57444558, -1.81113178,  2.96002293],
    #                         [ 0.60419802, -2.26035953,  3.46875557],
    #                         [ 0.63585684, -2.62746432,  3.899178  ],
    #                         [ 0.67238625, -2.94147506,  4.28624756],
    #                         [ 0.70992311, -3.22142148,  4.64126771],
    #                         [ 0.76089838, -3.46329933,  4.98509609],
    #                         [ 0.7855481 , -3.7121082 ,  5.28320441],
    #                         [ 0.73117064, -4.02408271,  5.48642398],
    #                         [ 0.669226  , -4.3302976 ,  5.66874961]]),
    #     index   = idx,
    #     columns = ['pred', 'lower_bound', 'upper_bound']
    # )

    # pd.testing.assert_frame_equal(preds_skforecast, expected)
    # pd.testing.assert_frame_equal(preds_statsmodels, expected)
    pd.testing.assert_frame_equal(preds_skforecast, preds_statsmodels)


def test_Sarimax_append_refit_False():
    """
    Test Sarimax append with refit `False`. It is tested with the 
    skforecast wrapper and the statsmodels SARIMAX.
    """
    
    refit = False

    # skforecast
    sarimax = Sarimax(order=(1, 1, 1), maxiter=1000, method='cg', disp=False)
    sarimax.fit(y=y_numpy, exog=exog_numpy)
    sarimax.append(
        y     = y_lw_numpy, 
        exog  = exog_lw_numpy,
        refit = refit
    )
    preds_skforecast = sarimax.predict(steps=10, exog=exog_lw_predict_numpy)

    # statsmodels
    sarimax = SARIMAX(order=(1, 1, 1), endog=y_numpy, exog=exog_numpy)               
    sarimax_res = sarimax.fit(maxiter=1000, method='cg', disp=False)
    updated_res = sarimax_res.append(
                      endog = y_lw_numpy, 
                      exog  = exog_lw_numpy,
                      refit = refit
                  )
    preds_statsmodels = updated_res.get_forecast(
                            steps = 10, 
                            exog  = exog_lw_predict_numpy
                        ).predicted_mean

    # expected = np.array([0.89386888, 0.92919515, 0.98327241, 1.02336286, 1.05334974,
    #                      1.08759298, 1.02651694, 0.97164917, 0.91539724, 0.85551674])

    # np.testing.assert_almost_equal(preds_skforecast, expected)
    # np.testing.assert_almost_equal(preds_statsmodels, expected)
    np.testing.assert_almost_equal(preds_skforecast, preds_statsmodels)


def test_Sarimax_append_refit_True():
    """
    Test Sarimax append with refit `True`. It is tested with the 
    skforecast wrapper and the statsmodels SARIMAX.
    """

    refit = True

    # skforecast
    sarimax = Sarimax(order=(1, 0, 1), maxiter=1000, method='cg', disp=False)
    sarimax.fit(y=y_numpy, exog=exog_numpy)
    sarimax.append(
        y     = y_lw_numpy, 
        exog  = exog_lw_numpy,
        refit = refit
    )
    preds_skforecast = sarimax.predict(steps=10, exog=exog_lw_predict_numpy)

    # statsmodels
    sarimax = SARIMAX(order=(1, 0, 1), endog=y_numpy, exog=exog_numpy)               
    sarimax_res = sarimax.fit(maxiter=1000, method='cg', disp=False)
    updated_res = sarimax_res.append(
                      endog = y_lw_numpy, 
                      exog  = exog_lw_numpy,
                      refit = refit,
                      fit_kwargs = {'method': 'cg', 
                                    'maxiter': 1000, 
                                    'start_params': None, 
                                    'disp': True}
                  )
    preds_statsmodels = updated_res.get_forecast(
                            steps = 10, 
                            exog  = exog_lw_predict_numpy
                        ).predicted_mean

    # statsmodels fitting all in the same array (not using append)
    sarimax = SARIMAX(
                  order = (1, 0, 1), 
                  endog = np.append(y_numpy, y_lw_numpy), 
                  exog  = np.append(exog_numpy, exog_lw_numpy)
              )       
    sarimax_res = sarimax.fit(maxiter=1000, method='cg', disp=False)
    preds_statsmodels_2 = updated_res.get_forecast(
                              steps = 10, 
                              exog  = exog_lw_predict_numpy
                          ).predicted_mean

    # expected = np.array([0.73036069, 0.72898841, 0.76386263, 0.78298827, 0.7990083 ,
    #                      0.81640187, 0.78540917, 0.75749361, 0.72887628, 0.69840139])

    # np.testing.assert_almost_equal(preds_skforecast, expected)
    # np.testing.assert_almost_equal(preds_statsmodels, expected)
    # np.testing.assert_almost_equal(preds_statsmodels_2, expected)
    np.testing.assert_almost_equal(preds_skforecast, preds_statsmodels)
    np.testing.assert_almost_equal(preds_skforecast, preds_statsmodels_2)


def test_Sarimax_apply_refit_False():
    """
    Test Sarimax apply with refit `False`. It is tested with the 
    skforecast wrapper and the statsmodels SARIMAX.
    """
    
    refit = False

    # skforecast
    sarimax = Sarimax(order=(1, 1, 1), maxiter=1000, method='cg', disp=False)
    sarimax.fit(y=y_numpy, exog=exog_numpy)
    sarimax.apply(
        y     = y_lw_numpy, 
        exog  = exog_lw_numpy,
        refit = refit
    )
    preds_skforecast = sarimax.predict(steps=10, exog=exog_lw_predict_numpy)

    # statsmodels
    sarimax = SARIMAX(order=(1, 1, 1), endog=y_numpy, exog=exog_numpy)               
    sarimax_res = sarimax.fit(maxiter=1000, method='cg', disp=False)
    updated_res = sarimax_res.apply(
                      endog = y_lw_numpy, 
                      exog  = exog_lw_numpy,
                      refit = refit
                  )
    preds_statsmodels = updated_res.get_forecast(
                            steps = 10, 
                            exog  = exog_lw_predict_numpy
                        ).predicted_mean

    # expected = np.array([0.89386888, 0.92919515, 0.98327241, 1.02336286, 1.05334974,
    #                      1.08759298, 1.02651694, 0.97164917, 0.91539724, 0.85551674])

    # np.testing.assert_almost_equal(preds_skforecast, expected)
    # np.testing.assert_almost_equal(preds_statsmodels, expected)
    np.testing.assert_almost_equal(preds_skforecast, preds_statsmodels)


def test_Sarimax_apply_refit_True():
    """
    Test Sarimax apply with refit `True`. It is tested with the 
    skforecast wrapper and the statsmodels SARIMAX.
    """

    refit = True

    # skforecast
    sarimax = Sarimax(order=(1, 0, 1), maxiter=1000, method='cg', disp=False)
    sarimax.fit(y=y_numpy, exog=exog_numpy)
    sarimax.apply(
        y     = y_lw_numpy, 
        exog  = exog_lw_numpy,
        refit = refit
    )
    preds_skforecast = sarimax.predict(steps=10, exog=exog_lw_predict_numpy)

    # statsmodels
    sarimax = SARIMAX(order=(1, 0, 1), endog=y_numpy, exog=exog_numpy)               
    sarimax_res = sarimax.fit(maxiter=1000, method='cg', disp=False)
    updated_res = sarimax_res.apply(
                      endog = y_lw_numpy, 
                      exog  = exog_lw_numpy,
                      refit = refit,
                      fit_kwargs = {'method': 'cg', 
                                    'maxiter': 1000, 
                                    'start_params': None, 
                                    'disp': True}
                  )
    preds_statsmodels = updated_res.get_forecast(
                            steps = 10, 
                            exog  = exog_lw_predict_numpy
                        ).predicted_mean

    # This is the same as creating a new SARIMAX and fitting with the new data
    sarimax = SARIMAX(
                  order = (1, 0, 1), 
                  endog = y_lw_numpy, 
                  exog  = exog_lw_numpy
              )       
    sarimax_res = sarimax.fit(maxiter=1000, method='cg', disp=False)
    preds_statsmodels_2 = updated_res.get_forecast(
                              steps = 10, 
                              exog  = exog_lw_predict_numpy
                          ).predicted_mean

    # expected = np.array([0.76718596, 0.75795261, 0.79506182, 0.81508184, 0.83164173,
    #                      0.84979957, 0.81751954, 0.78846941, 0.75867965, 0.72695935])

    # np.testing.assert_almost_equal(preds_skforecast, expected)
    # np.testing.assert_almost_equal(preds_statsmodels, expected)
    # np.testing.assert_almost_equal(preds_statsmodels_2, expected)
    np.testing.assert_almost_equal(preds_skforecast, preds_statsmodels)
    np.testing.assert_almost_equal(preds_skforecast, preds_statsmodels_2)


def test_Sarimax_extend():
    """
    Test Sarimax extend. It is tested with the skforecast wrapper and the 
    statsmodels SARIMAX.
    """

    # skforecast
    sarimax = Sarimax(order=(1, 1, 1), maxiter=1000, method='cg', disp=False)
    sarimax.fit(y=y_numpy, exog=exog_numpy)
    sarimax.extend(
        y     = y_lw_numpy, 
        exog  = exog_lw_numpy
    )
    preds_skforecast = sarimax.predict(steps=10, exog=exog_lw_predict_numpy)

    # statsmodels
    sarimax = SARIMAX(order=(1, 1, 1), endog=y_numpy, exog=exog_numpy)
    sarimax_res = sarimax.fit(maxiter=1000, method='cg', disp=False)
    updated_res = sarimax_res.extend(
                      endog = y_lw_numpy, 
                      exog  = exog_lw_numpy
                  )
    preds_statsmodels = updated_res.get_forecast(
                            steps = 10, 
                            exog  = exog_lw_predict_numpy
                        ).predicted_mean

    # expected = np.array([0.89386888, 0.92919515, 0.98327241, 1.02336286, 1.05334974,
    #                      1.08759298, 1.02651694, 0.97164917, 0.91539724, 0.85551674])

    # np.testing.assert_almost_equal(preds_skforecast, expected)
    # np.testing.assert_almost_equal(preds_statsmodels, expected)
    np.testing.assert_almost_equal(preds_skforecast, preds_statsmodels)


def test_Sarimax_set_params():
    """
    Test Sarimax set_params.
    """
    
    sarimax = Sarimax(order=(1, 1, 1))

    sarimax.set_params(
        order         = (2, 2, 2),
        maxiter       = 200,
        sm_fit_kwargs = {'test': 1},
        not_a_param   = 'fake'
    )

    results = sarimax.get_params()

    expected = {
        'order': (2, 2, 2),
        'seasonal_order': (0, 0, 0, 0),
        'trend': None,
        'measurement_error': False,
        'time_varying_regression': False,
        'mle_regression': True,
        'simple_differencing': False,
        'enforce_stationarity': True,
        'enforce_invertibility': True,
        'hamilton_representation': False,
        'concentrate_scale': False,
        'trend_offset': 1,
        'use_exact_diffuse': False,
        'dates': None,
        'freq': None,
        'missing': 'none',
        'validate_specification': True,
        'method': 'lbfgs',
        'maxiter': 200,
        'start_params': None,
        'disp': False,
        'sm_init_kwargs': {},
        'sm_fit_kwargs': {'test': 1},
        'sm_predict_kwargs': {}
    }

    assert results == expected
    assert sarimax.output_type is None
    assert sarimax.sarimax_res is None
    assert sarimax.fitted == False
    assert sarimax.training_index is None


def test_Sarimax_params_numpy():
    """
    Test Sarimax params after fit with numpy `y` and `exog`.
    """

    sarimax_sk = Sarimax(order=(3, 2, 0), disp=0)
    sarimax_sk.fit(y=y_numpy, exog=exog_numpy)

    sarimax_sm = SARIMAX(order=(3, 2, 0), endog=y_numpy, exog=exog_numpy)
    sarimax_res = sarimax_sm.fit(disp=0)

    results_skforecast = sarimax_sk.params()
    results_statsmodels = sarimax_res.params

    # expected = np.array([0.46809671, 0.24916836, 0.45027591, 0.0105501 ])

    # np.testing.assert_almost_equal(results_skforecast, expected)
    # np.testing.assert_almost_equal(results_statsmodels, expected)
    np.testing.assert_almost_equal(results_skforecast, results_statsmodels)


@pytest.mark.parametrize("y, exog", 
                         [(y, exog),
                          (y_datetime, exog_datetime),
                          (y.to_frame(), exog.to_frame()),
                          (y_datetime.to_frame(), exog_datetime.to_frame())], 
                         ids = lambda values : f'y, exog: {type(values)}')
def test_Sarimax_params_pandas(y, exog):
    """
    Test Sarimax params after fit with pandas `y` and `exog`.
    """

    sarimax_sk = Sarimax(order=(3, 2, 0), disp=0)
    sarimax_sk.fit(y=y, exog=exog)

    sarimax_sm = SARIMAX(order=(3, 2, 0), endog=y, exog=exog)
    sarimax_res = sarimax_sm.fit(disp=0)

    results_skforecast = sarimax_sk.params()
    results_statsmodels = sarimax_res.params

    # expected = pd.Series(
    #                data  = np.array([0.46809671, 0.24916836, 0.45027591, 0.0105501 ]),
    #                index = ['exog', 'ar.L1', 'ma.L1', 'sigma2'],
    #            )

    # pd.testing.assert_series_equal(results_skforecast, expected)
    # pd.testing.assert_series_equal(results_statsmodels, expected)
    pd.testing.assert_series_equal(results_skforecast, results_statsmodels)


@pytest.mark.parametrize("y, exog", 
                         [(y_numpy, exog_numpy),
                          (y, exog),
                          (y_datetime, exog_datetime),
                          (y.to_frame(), exog.to_frame()),
                          (y_datetime.to_frame(), exog_datetime.to_frame())], 
                         ids = lambda values : f'y, exog: {type(values)}')
def test_Sarimax_summary(y, exog):
    """
    Test Sarimax params after fit with pandas `y` and `exog`.
    """

    sarimax = Sarimax(order=(1, 0, 1))
    sarimax.fit(y=y, exog=exog)

    # Only 316 characters are checked because the summary includes the date and time of the training
    results = sarimax.summary().as_text()[:316]
    expected = (
        '                               SARIMAX Results                                \n'
        '==============================================================================\n'
        'Dep. Variable:                      y   No. Observations:                   50\n'
        'Model:               SARIMAX(1, 0, 1)   Log Likelihood                  42.591\n'
    )
    
    assert results == expected


def test_Sarimax_get_info_criteria_ValueError_criteria_invalid_value():
    """
    Test Sarimax get_info_criteria ValueError when `criteria` is an invalid value.
    """

    sarimax = Sarimax(order=(1, 0, 1))
    sarimax.fit(y=y, exog=exog)

    criteria = 'not_valid'

    err_msg = re.escape(
        ("Invalid value for `criteria`. Valid options are 'aic', 'bic', "
         "and 'hqic'.")
    )
    with pytest.raises(ValueError, match = err_msg): 
        sarimax.get_info_criteria(criteria=criteria)


def test_Sarimax_get_info_criteria_ValueError_method_invalid_value():
    """
    Test Sarimax get_info_criteria ValueError when `method` is an invalid value.
    """

    sarimax = Sarimax(order=(1, 0, 1))
    sarimax.fit(y=y, exog=exog)

    method = 'not_valid'

    err_msg = re.escape(
        ("Invalid value for `method`. Valid options are 'standard' and "
         "'lutkepohl'.")
    )
    with pytest.raises(ValueError, match = err_msg): 
        sarimax.get_info_criteria(method=method)


@pytest.mark.parametrize("y, exog", 
                         [(y_numpy, exog_numpy),
                          (y, exog),
                          (y_datetime, exog_datetime),
                          (y.to_frame(), exog.to_frame()),
                          (y_datetime.to_frame(), exog_datetime.to_frame())], 
                         ids = lambda values : f'y, exog: {type(values)}')
def test_Sarimax_get_info_criteria(y, exog):
    """
    Test Sarimax get_info_criteria after fit `y` and `exog`.
    """
    
    sarimax_sk = Sarimax(order=(3, 2, 0), disp=0)
    sarimax_sk.fit(y=y, exog=exog)

    sarimax_sm = SARIMAX(order=(3, 2, 0), endog=y, exog=exog)
    sarimax_res = sarimax_sm.fit(disp=0)

    results_skforecast = sarimax_sk.get_info_criteria(criteria='aic', method='standard')
    results_statsmodels = sarimax_res.info_criteria(criteria='aic', method='standard')

    # expected = -42.00040332298444

    # assert results_skforecast == expected
    # assert results_statsmodels == expected
    assert results_skforecast == pytest.approx(results_statsmodels)


# Unit test predict_bootstrapping ForecasterAutoregMultiSeriesCustom
# ==============================================================================
import re
import pytest
import numpy as np
import pandas as pd
from sklearn.exceptions import NotFittedError
from sklearn.compose import ColumnTransformer
from sklearn.preprocessing import StandardScaler
from sklearn.preprocessing import OneHotEncoder
from sklearn.linear_model import LinearRegression

from skforecast.exceptions import IgnoredArgumentWarning
from skforecast.ForecasterAutoregMultiSeriesCustom import ForecasterAutoregMultiSeriesCustom

# Fixtures
from .fixtures_ForecasterAutoregMultiSeriesCustom import series
from .fixtures_ForecasterAutoregMultiSeriesCustom import exog
from .fixtures_ForecasterAutoregMultiSeriesCustom import exog_predict

transformer_exog = ColumnTransformer(
                       [('scale', StandardScaler(), ['exog_1']),
                        ('onehot', OneHotEncoder(), ['exog_2'])],
                       remainder = 'passthrough',
                       verbose_feature_names_out = False
                   )

def create_predictors(y): # pragma: no cover
    """
    Create first 3 lags of a time series.
    """
    lags = y[-1:-4:-1]

    return lags


def test_predict_bootstrapping_NotFittedError_when_fitted_is_False():
    """
    Test NotFittedError is raised when fitted is False.
    """
    forecaster = ForecasterAutoregMultiSeriesCustom(
                     regressor      = LinearRegression(),
                     fun_predictors = create_predictors,
                     window_size    = 3
                 )

    err_msg = re.escape(
        ("This Forecaster instance is not fitted yet. Call `fit` with "
         "appropriate arguments before using predict.")
    )
    with pytest.raises(NotFittedError, match = err_msg):
        forecaster.predict_bootstrapping(steps=1, levels=None)


def test_predict_bootstrapping_IgnoredArgumentWarning_when_not_available_self_last_window_for_some_levels():
    """
    Test IgnoredArgumentWarning is raised when last_window is not available for 
    levels because it was not stored during fit.
    """
    forecaster = ForecasterAutoregMultiSeriesCustom(
                     regressor          = LinearRegression(),
                     fun_predictors     = create_predictors,
                     window_size        = 3,
                     transformer_series = None
                 )
    forecaster.fit(series=series, exog=exog['exog_1'], store_last_window=['1'])

    warn_msg = re.escape(
        ("Levels {'2'} are excluded from prediction "
         "since they were not stored in `last_window` attribute "
         "during training. If you don't want to retrain the "
         "Forecaster, provide `last_window` as argument.")
    )
    with pytest.warns(IgnoredArgumentWarning, match = warn_msg):
        results = forecaster.predict_bootstrapping(
                      steps               = 1,
                      levels              = ['1', '2'],
                      n_boot              = 4, 
                      exog                = exog_predict['exog_1'], 
                      in_sample_residuals = True
                  )

    expected_1 = pd.DataFrame(
                     data    = np.array([[0.16386591, 0.56398143, 0.38576231, 0.37782729]]),
                     columns = [f"pred_boot_{i}" for i in range(4)],
                     index   = pd.RangeIndex(start=50, stop=51)
                 )

    expected = {'1': expected_1 }

    pd.testing.assert_frame_equal(results['1'], expected['1'])


@pytest.mark.parametrize("store_last_window",
                         [['1'], False],
                         ids=lambda slw: f"store_last_window: {slw}")
def test_predict_bootstrapping_ValueError_when_not_available_self_last_window_for_levels(store_last_window):
    """
    Test ValueError is raised when last_window is not available for all 
    levels because it was not stored during fit.
    """
    forecaster = ForecasterAutoregMultiSeriesCustom(
                     regressor          = LinearRegression(),
                     fun_predictors     = create_predictors,
                     window_size        = 3,
                     transformer_series = None
                 )
    forecaster.fit(series=series, store_last_window=store_last_window)

    err_msg = re.escape(
        ("No series to predict. None of the series are present in "
         "`last_window` attribute. Provide `last_window` as argument "
         "in predict method.")
    )
    with pytest.raises(ValueError, match = err_msg):
        forecaster.predict_bootstrapping(steps=5, levels=['2'], last_window=None)


def test_predict_bootstrapping_IgnoredArgumentWarning_when_levels_is_list_and_different_last_index_in_self_last_window_DatetimeIndex():
    """
    Test IgnoredArgumentWarning is raised when levels is a list and have 
    different last index in last_window attribute using a DatetimeIndex.
    """
    series_2 = {
        '1': series['1'].copy(),
        '2': series['2'].iloc[:30].copy()
    }
    series_2['1'].index = pd.date_range(start='2020-01-01', periods=50)
    series_2['2'].index = pd.date_range(start='2020-01-01', periods=30)
    exog_2 = {
        '1': exog['exog_1'].copy(),
        '2': exog['exog_1'].iloc[:30].copy()
    }
    exog_2['1'].index = pd.date_range(start='2020-01-01', periods=50)
    exog_2['2'].index = pd.date_range(start='2020-01-01', periods=30)
    exog_2_pred = {
        '1': exog_predict['exog_1'].copy()
    }
    exog_2_pred['1'].index = pd.date_range(start='2020-02-20', periods=50)

    forecaster = ForecasterAutoregMultiSeriesCustom(
                     regressor          = LinearRegression(),
                     fun_predictors     = create_predictors,
                     window_size        = 3,
                     transformer_series = None
                 )
    forecaster.fit(series=series_2, exog=exog_2)

    warn_msg = re.escape(
        ("Only series whose last window ends at the same index "
         "can be predicted together. Series that do not reach the "
         "maximum index, '2020-02-19 00:00:00', are excluded "
         "from prediction: {'2'}.")
    )
    with pytest.warns(IgnoredArgumentWarning, match = warn_msg):
        results = forecaster.predict_bootstrapping(
                      steps               = 1,
                      levels              = ['1', '2'],
                      n_boot              = 4, 
                      exog                = exog_2_pred, 
                      in_sample_residuals = True
                  )

    expected_1 = pd.DataFrame(
                     data    = np.array([[0.17482136, 0.59548022, 0.40159385, 0.37372285]]),
                     columns = [f"pred_boot_{i}" for i in range(4)],
                     index   = pd.date_range(start='2020-02-20', periods=1)
                 )

    expected = {'1': expected_1 }

    pd.testing.assert_frame_equal(results['1'], expected['1'])


def test_predict_bootstrapping_ValueError_when_not_in_sample_residuals_for_any_level():
    """
    Test ValueError is raised when in_sample_residuals=True but there is no
    residuals for any level.
    """
    forecaster = ForecasterAutoregMultiSeriesCustom(
                     regressor       = LinearRegression(),
                     fun_predictors  = create_predictors,
                     window_size     = 3
                 )
    
    forecaster.fit(series=series)
    forecaster.in_sample_residuals = {2: np.array([1, 2, 3])}

    err_msg = re.escape(
                (f"Not `forecaster.in_sample_residuals` for levels: "
                 f"{set(['1', '2']) - set(forecaster.in_sample_residuals.keys())}.")
                )
    with pytest.raises(ValueError, match = err_msg):
        forecaster.predict_bootstrapping(steps=1, levels=None, in_sample_residuals=True)


def test_predict_bootstrapping_ValueError_when_out_sample_residuals_is_None():
    """
    Test ValueError is raised when in_sample_residuals=False and
    forecaster.out_sample_residuals is None.
    """
    forecaster = ForecasterAutoregMultiSeriesCustom(
                     regressor       = LinearRegression(),
                     fun_predictors  = create_predictors,
                     window_size     = 3
                 )
    forecaster.fit(series=series)

    err_msg = re.escape(
        ("`forecaster.out_sample_residuals` is `None`. Use "
         "`in_sample_residuals=True` or method "
         "`set_out_sample_residuals()` before `predict_interval()`, "
         "`predict_bootstrapping()`,`predict_quantiles()` or "
         "`predict_dist()`.")
    )
    with pytest.raises(ValueError, match = err_msg):
        forecaster.predict_bootstrapping(steps=1, levels='1', in_sample_residuals=False)


def test_predict_bootstrapping_ValueError_when_not_out_sample_residuals_for_all_levels():
    """
    Test ValueError is raised when in_sample_residuals=False and
    forecaster.out_sample_residuals is not available for all levels.
    """
    forecaster = ForecasterAutoregMultiSeriesCustom(
                     regressor       = LinearRegression(),
                     fun_predictors  = create_predictors,
                     window_size     = 3
                 )
    forecaster.fit(series=series)
    residuals = {'2': np.array([1, 2, 3, 4, 5]), 
                 '3': np.array([1, 2, 3, 4, 5])}
    forecaster.out_sample_residuals = residuals
    levels = ['1']

    err_msg = re.escape(
        (f"Not `forecaster.out_sample_residuals` for levels: "
         f"{set(levels) - set(forecaster.out_sample_residuals.keys())}. "
         f"Use method `set_out_sample_residuals()`.")
    )
    with pytest.raises(ValueError, match = err_msg):
        forecaster.predict_bootstrapping(steps=1, levels=levels, in_sample_residuals=False)


@pytest.mark.parametrize("transformer_series", 
                         [None, StandardScaler()],
                         ids = lambda tr : f'transformer_series type: {type(tr)}')
def test_predict_bootstrapping_ValueError_when_level_out_sample_residuals_value_is_None(transformer_series):
    """
    Test ValueError is raised when in_sample_residuals=False and
    forecaster.out_sample_residuals has a level with a None.
    """
    forecaster = ForecasterAutoregMultiSeriesCustom(
        regressor=LinearRegression(),
        fun_predictors=create_predictors,
        window_size=3,
        transformer_series=transformer_series,
    )
    forecaster.fit(series=series)
    residuals = {'1': np.array([1, 2, 3, 4, 5])}
    forecaster.set_out_sample_residuals(residuals = residuals)

    err_msg = re.escape(
        ("Not available residuals for level '2'. "
         "Check `forecaster.out_sample_residuals`.")
    )
    with pytest.raises(ValueError, match = err_msg):
        forecaster.predict_bootstrapping(steps=3, in_sample_residuals=False)


@pytest.mark.parametrize("transformer_series", 
                         [None, StandardScaler()],
                         ids = lambda tr : f'transformer_series type: {type(tr)}')
def test_predict_bootstrapping_ValueError_when_level_out_sample_residuals_value_contains_None_or_NaNs(transformer_series):
    """
    Test ValueError is raised when in_sample_residuals=False and
    forecaster.out_sample_residuals has a level with a None or NaN value.
    """
    forecaster = ForecasterAutoregMultiSeriesCustom(
                     regressor          = LinearRegression(),
                     fun_predictors     = create_predictors,
                     window_size        = 3,
                     transformer_series = transformer_series
                 )
    forecaster.fit(series=series)
    residuals = {'1': np.array([1, 2, 3, 4, 5]),
                 '2': np.array([1, 2, 3, 4, None])} # StandardScaler() transforms None to NaN
    forecaster.set_out_sample_residuals(residuals = residuals)

    err_msg = re.escape(
        ("forecaster residuals for level '2' contains `None` "
         "or `NaNs` values. Check `forecaster.out_sample_residuals`.")
    )
    with pytest.raises(ValueError, match = err_msg):
        forecaster.predict_bootstrapping(steps=3, in_sample_residuals=False)


def test_predict_bootstrapping_output_when_forecaster_is_LinearRegression_exog_steps_is_1_in_sample_residuals_is_True():
    """
    Test output of predict_bootstrapping when regressor is LinearRegression and
    1 step ahead is predicted with exog using in-sample residuals.
    """
    forecaster = ForecasterAutoregMultiSeriesCustom(
                     regressor          = LinearRegression(),
                     fun_predictors     = create_predictors,
                     window_size        = 3,
                     transformer_series = None
                 )
    forecaster.fit(series=series, exog=exog['exog_1'])
    results = forecaster.predict_bootstrapping(
                  steps               = 1, 
                  n_boot              = 4, 
                  exog                = exog_predict['exog_1'], 
                  in_sample_residuals = True
              )

    expected_1 = pd.DataFrame(
                     data    = np.array([[0.16386591, 0.56398143, 0.38576231, 0.37782729]]),
                     columns = [f"pred_boot_{i}" for i in range(4)],
                     index   = pd.RangeIndex(start=50, stop=51)
                 )
    expected_2 = pd.DataFrame(
                     data    = np.array([[0.24898667, 0.69862572, 0.72572763, 0.45672811]]),
                     columns = [f"pred_boot_{i}" for i in range(4)],
                     index   = pd.RangeIndex(start=50, stop=51)
                 )

    expected = {'1': expected_1 ,'2': expected_2}

    for key in results.keys():
        pd.testing.assert_frame_equal(results[key], expected[key])
    
    
def test_predict_bootstrapping_output_when_forecaster_is_LinearRegression_exog_steps_is_2_in_sample_residuals_is_True():
    """
    Test output of predict_bootstrapping when regressor is LinearRegression and
    2 steps ahead are predicted with exog using in-sample residuals.
    """
    forecaster = ForecasterAutoregMultiSeriesCustom(
                     regressor          = LinearRegression(),
                     fun_predictors     = create_predictors,
                     window_size        = 3,
                     transformer_series = None
                 )
    forecaster.fit(series=series, exog=exog['exog_1'])
    results = forecaster.predict_bootstrapping(
                  steps               = 2, 
                  n_boot              = 4, 
                  exog                = exog_predict['exog_1'], 
                  in_sample_residuals = True
              )

    expected_1 = pd.DataFrame(
                     data    = np.array([[0.16386591, 0.56398143, 0.38576231, 0.37782729],
                                         [0.04906615, 0.45019329, 0.07478854, 0.222437  ]]),
                     columns = [f"pred_boot_{i}" for i in range(4)],
                     index   = pd.RangeIndex(start=50, stop=52)
                 )
    expected_2 = pd.DataFrame(
                     data    = np.array([[0.24898667, 0.69862572, 0.72572763, 0.45672811],
                                         [0.29374176, 0.40758093, 0.74577937, 0.11883889]]),
                     columns = [f"pred_boot_{i}" for i in range(4)],
                     index   = pd.RangeIndex(start=50, stop=52)
                 )    

    expected = {'1': expected_1 ,'2': expected_2}

    for key in results.keys():
        pd.testing.assert_frame_equal(results[key], expected[key])
    
    
def test_predict_bootstrapping_output_when_forecaster_is_LinearRegression_exog_steps_is_1_in_sample_residuals_is_False():
    """
    Test output of predict_bootstrapping when regressor is LinearRegression and
    1 step ahead is predicted with exog using out-sample residuals.
    """
    forecaster = ForecasterAutoregMultiSeriesCustom(
                     regressor          = LinearRegression(),
                     fun_predictors     = create_predictors,
                     window_size        = 3,
                     transformer_series = None
                 )
    forecaster.fit(series=series, exog=exog['exog_1'])
    forecaster.out_sample_residuals = forecaster.in_sample_residuals
    results = forecaster.predict_bootstrapping(
                  steps               = 1, 
                  n_boot              = 4, 
                  exog                = exog_predict['exog_1'], 
                  in_sample_residuals = False,
                  suppress_warnings   = True
              )

    expected_1 = pd.DataFrame(
                     data    = np.array([[0.16386591, 0.56398143, 0.38576231, 0.37782729]]),
                     columns = [f"pred_boot_{i}" for i in range(4)],
                     index   = pd.RangeIndex(start=50, stop=51)
                 )
    expected_2 = pd.DataFrame(
                     data    = np.array([[0.24898667, 0.69862572, 0.72572763, 0.45672811]]),
                     columns = [f"pred_boot_{i}" for i in range(4)],
                     index   = pd.RangeIndex(start=50, stop=51)
                 )

    expected = {'1': expected_1 ,'2': expected_2}

    for key in results.keys():
        pd.testing.assert_frame_equal(results[key], expected[key])
    
    
def test_predict_bootstrapping_output_when_forecaster_is_LinearRegression_exog_steps_is_2_in_sample_residuals_is_False():
    """
    Test output of predict_bootstrapping when regressor is LinearRegression and
    2 steps ahead are predicted with exog using out-sample residuals.
    """
    forecaster = ForecasterAutoregMultiSeriesCustom(
                     regressor          = LinearRegression(),
                     fun_predictors     = create_predictors,
                     window_size        = 3,
                     transformer_series = None
                 )
    forecaster.fit(series=series, exog=exog['exog_1'])
    forecaster.out_sample_residuals = forecaster.in_sample_residuals
    results = forecaster.predict_bootstrapping(
                  steps               = 2, 
                  n_boot              = 4, 
                  exog                = exog_predict['exog_1'], 
                  in_sample_residuals = False
              )
    
    expected_1 = pd.DataFrame(
                     data    = np.array([[0.16386591, 0.56398143, 0.38576231, 0.37782729],
                                         [0.04906615, 0.45019329, 0.07478854, 0.222437  ]]),
                     columns = [f"pred_boot_{i}" for i in range(4)],
                     index   = pd.RangeIndex(start=50, stop=52)
                 )
    expected_2 = pd.DataFrame(
                     data    = np.array([[0.24898667, 0.69862572, 0.72572763, 0.45672811],
                                         [0.29374176, 0.40758093, 0.74577937, 0.11883889]]),
                     columns = [f"pred_boot_{i}" for i in range(4)],
                     index   = pd.RangeIndex(start=50, stop=52)
                 )  

    expected = {'1': expected_1 ,'2': expected_2}

    for key in results.keys():
        pd.testing.assert_frame_equal(results[key], expected[key])


def test_predict_bootstrapping_output_when_forecaster_is_LinearRegression_steps_is_2_in_sample_residuals_True_exog_and_transformer_fixed_to_zero():
    """
    Test output of predict_bootstrapping when regressor is LinearRegression,
    2 steps are predicted, using in-sample residuals, exog is included, both
    inputs are transformed and in-sample residuals are fixed to 0.
    """    
    forecaster = ForecasterAutoregMultiSeriesCustom(
                     regressor          = LinearRegression(),
                     fun_predictors     = create_predictors,
                     window_size        = 3,
                     transformer_series = StandardScaler(),
                     transformer_exog   = transformer_exog,
                 )
    forecaster.fit(series=series, exog=exog)

    forecaster.in_sample_residuals['1'] = np.full_like(forecaster.in_sample_residuals['1'], fill_value=0)
    forecaster.in_sample_residuals['2'] = np.full_like(forecaster.in_sample_residuals['2'], fill_value=0)
    results = forecaster.predict_bootstrapping(
                  steps               = 2, 
                  n_boot              = 4, 
                  exog                = exog_predict, 
                  in_sample_residuals = True
              )
    
    expected_1 = pd.DataFrame(
                     data    = np.array([[0.50201669, 0.50201669, 0.50201669, 0.50201669],
                                         [0.49804821, 0.49804821, 0.49804821, 0.49804821]]),
                     columns = [f"pred_boot_{i}" for i in range(4)],
                     index   = pd.RangeIndex(start=50, stop=52)
                 )
    expected_2 = pd.DataFrame(
                     data    = np.array([[0.52531076, 0.52531076, 0.52531076, 0.52531076],
                                         [0.51683613, 0.51683613, 0.51683613, 0.51683613]]),
                     columns = [f"pred_boot_{i}" for i in range(4)],
                     index   = pd.RangeIndex(start=50, stop=52)
                 )  

    expected = {'1': expected_1 ,'2': expected_2}

    for key in results.keys():
        pd.testing.assert_frame_equal(results[key], expected[key])


def test_predict_bootstrapping_output_when_forecaster_is_LinearRegression_steps_is_2_in_sample_residuals_True_exog_and_transformer():
    """
    Test output of predict_bootstrapping when regressor is LinearRegression,
    2 steps are predicted, using in-sample residuals, exog is included and both
    inputs are transformed.
    """
    forecaster = ForecasterAutoregMultiSeriesCustom(
                     regressor          = LinearRegression(),
                     fun_predictors     = create_predictors,
                     window_size        = 3,
                     transformer_series = StandardScaler(),
                     transformer_exog   = transformer_exog,
                 )
    forecaster.fit(series=series, exog=exog)
    results = forecaster.predict_bootstrapping(
                  steps               = 2, 
                  n_boot              = 4, 
                  exog                = exog_predict, 
                  in_sample_residuals = True
              )
    
    expected_1 = pd.DataFrame(
                     data    = np.array([[0.16958001, 0.55887507, 0.41716867, 0.38128471],
                                         [0.08114962, 0.48871107, 0.07673561, 0.25559782]]),
                     columns = [f"pred_boot_{i}" for i in range(4)],
                     index   = pd.RangeIndex(start=50, stop=52)
                 )
    expected_2 = pd.DataFrame(
                     data    = np.array([[0.24448676, 0.70066373, 0.77244769, 0.45857602],
                                         [0.33048491, 0.44364526, 0.75067101, 0.16052538]]),
                     columns = [f"pred_boot_{i}" for i in range(4)],
                     index   = pd.RangeIndex(start=50, stop=52)
                 )  

    expected = {'1': expected_1 ,'2': expected_2}

    for key in results.keys():
        pd.testing.assert_frame_equal(results[key], expected[key])


# Unit test check_preprocess_series
# ==============================================================================
import re
import pytest
import numpy as np
import pandas as pd
from skforecast.utils import check_preprocess_series
from skforecast.ForecasterAutoregMultiSeries.tests.fixtures_ForecasterAutoregMultiSeries import series
from skforecast.ForecasterAutoregMultiSeries.tests.fixtures_ForecasterAutoregMultiSeries import series_as_dict


def test_TypeError_check_preprocess_series_when_series_is_not_pandas_DataFrame_or_dict():
    """
    Test TypeError is raised when series is not a pandas DataFrame or dict.
    """
    series = np.array([1, 2, 3])

    err_msg = re.escape(
        (f"`series` must be a pandas DataFrame or a dict of DataFrames or Series. "
         f"Got {type(series)}.")
    )
    with pytest.raises(TypeError, match = err_msg):
        check_preprocess_series(series = series)


def test_ValueError_check_preprocess_series_when_all_series_values_are_missing_DataFrame():
    """
    Test ValueError is raised when all series values are missing when series
    is a pandas DataFrame.
    """
    series_nan = pd.DataFrame({'1': pd.Series(np.arange(7)), 
                               '2': pd.Series([np.nan]*7)})
    series_nan.index = pd.date_range(start='2022-01-01', periods=7, freq='1D')

    err_msg = re.escape("All values of series '2' are NaN.")
    with pytest.raises(ValueError, match = err_msg):
        check_preprocess_series(series=series_nan)


def test_check_preprocess_series_when_series_is_pandas_DataFrame():
    """
    Test check_preprocess_series when `series` is a pandas DataFrame.
    """
    series_dict, series_indexes = check_preprocess_series(series=series)

    expected_series_indexes = {
        col: pd.RangeIndex(start=0, stop=len(series), step=1) 
        for col in series.columns
    }

    assert isinstance(series_dict, dict)
    assert list(series_dict.keys()) == ['1', '2']
    for k, v in series_dict.items():
        assert isinstance(v, pd.Series)
        assert k == v.name

    assert isinstance(series_indexes, dict)
    assert list(series_indexes.keys()) == ['1', '2']
    for k in series_indexes:
        pd.testing.assert_index_equal(series_indexes[k], expected_series_indexes[k])
        assert series_indexes[k].step == expected_series_indexes[k].step


def test_check_preprocess_series_when_series_is_pandas_DataFrame_with_DatetimeIndex():
    """
    Test check_preprocess_series when `series` is a pandas DataFrame with 
    a DatetimeIndex.
    """
    series_datetime = series.copy()
    series_datetime.index = pd.date_range(
        start='2000-01-01', periods=len(series_datetime), freq='D'
    )

    series_dict, series_indexes = check_preprocess_series(series=series_datetime)

    expected_series_indexes = {series: pd.date_range(
        start='2000-01-01', periods=len(series_datetime), freq='D'
    ) for series in series_datetime.columns}

    assert isinstance(series_dict, dict)
    assert list(series_dict.keys()) == ['1', '2']
    for k, v in series_dict.items():
        assert isinstance(v, pd.Series)
        assert k == v.name
        pd.testing.assert_series_equal(v, series_datetime[k])

    assert isinstance(series_indexes, dict)
    assert list(series_indexes.keys()) == ['1', '2']
    for k in series_indexes:
        pd.testing.assert_index_equal(series_indexes[k], expected_series_indexes[k])
        assert series_indexes[k].freq == expected_series_indexes[k].freq


def test_check_preprocess_series_when_series_is_pandas_DataFrame_with_no_pandas_Index():
    """
    Test check_preprocess_series when `series` is a pandas DataFrame with
    no pandas Index.
    """
    series_no_pandas_idx = series.copy()
    series_no_pandas_idx.index = np.arange(6, 6 + len(series))
    series_dict, series_indexes = check_preprocess_series(series=series_no_pandas_idx)

    expected_series_indexes = {
        col: pd.RangeIndex(start=0, stop=len(series_no_pandas_idx), step=1) 
        for col in series_no_pandas_idx.columns
    }

    assert isinstance(series_dict, dict)
    assert list(series_dict.keys()) == ['1', '2']
    for k, v in series_dict.items():
        assert isinstance(v, pd.Series)
        assert k == v.name

    assert isinstance(series_indexes, dict)
    assert list(series_indexes.keys()) == ['1', '2']
    for k in series_indexes:
        pd.testing.assert_index_equal(series_indexes[k], expected_series_indexes[k])
        assert series_indexes[k].step == expected_series_indexes[k].step


def test_TypeError_check_preprocess_series_when_series_is_dict_with_no_pandas_Series_or_DataFrame():
    """
    Test TypeError is raised when series is dict not containing 
    a pandas Series or DataFrame.
    """
    series_dict = {'l1': np.array([1, 2, 3]),
                   'l2': series}

    err_msg = re.escape(
        ("If `series` is a dictionary, all series must be a named pandas Series "
         "or a pandas DataFrame with a single column. Review series: ['l1']")
    )
    with pytest.raises(TypeError, match = err_msg):
        check_preprocess_series(series = series_dict)


def test_ValueError_check_preprocess_series_when_series_is_dict_with_a_DataFrame_with_2_columns():
    """
    Test ValueError is raised when series is dict containing a pandas DataFrame 
    with 2 columns.
    """
    series_dict = {'l1': series['1'],
                   'l2': series}

    err_msg = re.escape(
        ("If `series` is a dictionary, all series must be a named pandas Series "
         "or a pandas DataFrame with a single column. Review series: 'l2'")
    )
    with pytest.raises(ValueError, match = err_msg):
        check_preprocess_series(series = series_dict)


def test_TypeError_check_preprocess_series_when_series_is_dict_with_no_DatetimeIndex():
    """
    Test TypeError is raised when series is dict containing series with no
    DatetimeIndex.
    """
    series_dict = {'l1': series['1'],
                   'l2': series['2']}

    err_msg = re.escape(
        ("If `series` is a dictionary, all series must have a Pandas DatetimeIndex "
         "as index with the same frequency. Review series: ['l1', 'l2']")
    )
    with pytest.raises(TypeError, match = err_msg):
        check_preprocess_series(series = series_dict)


def test_ValueError_check_preprocess_series_when_series_is_dict_with_different_freqs():
    """
    Test ValueError is raised when series is dict containing series with 
    DatetimeIndex but different frequencies.
    """
    series_dict = {
        'l1': series_as_dict['l1'].copy(),
        'l2': series_as_dict['l2'].copy()
    }
    series_dict['l1'].index = pd.date_range(
        start='2000-01-01', periods=len(series_dict['l1']), freq='D'
    )
    series_dict['l2'].index = pd.date_range(
        start='2000-01-01', periods=len(series_dict['l2']), freq='MS'
    )

    err_msg = re.escape(
        ("If `series` is a dictionary, all series must have a Pandas DatetimeIndex "
         "as index with the same frequency. Found frequencies: ['<Day>', '<MonthBegin>']")
    )
    with pytest.raises(ValueError, match = err_msg):
        check_preprocess_series(series = series_dict)


def test_ValueError_check_preprocess_series_when_all_series_values_are_missing_dict():
    """
    Test ValueError is raised when all series values are missing when series
    is a dict.
    """
    series_nan = pd.DataFrame({'l1': pd.Series(np.arange(7)), 
                               'l2': pd.Series([np.nan]*7)})
    series_nan.index = pd.date_range(start='2022-01-01', periods=7, freq='1D')
    series_nan = series_nan.to_dict("series")

    err_msg = re.escape("All values of series 'l2' are NaN.")
    with pytest.raises(ValueError, match = err_msg):
        check_preprocess_series(series=series_nan)


def test_check_preprocess_series_when_series_is_dict():
    """
    Test check_preprocess_series when `series` is a pandas DataFrame with 
    a DatetimeIndex.
    """
    series_dict = {
        'l1': series_as_dict['l1'].copy(),
        'l2': series_as_dict['l2'].copy()
    }

    series_dict['l1'].index = pd.date_range(start='2000-01-01', periods=len(series_dict['l1']), freq='D')
    series_dict['l2'] = series_dict['l2'].to_frame()
    series_dict['l2'].index = pd.date_range(start='2000-01-01', periods=len(series_dict['l2']), freq='D')

    series_dict, series_indexes = check_preprocess_series(series=series_dict)

    expected_series_dict = {
        'l1': pd.Series(
                  data  = series_dict['l1'].to_numpy(),
                  index = pd.date_range(start='2000-01-01', periods=len(series['1']), freq='D'),
                  name  ='l1'
              ),
        'l2': pd.Series(
                  data  = series_dict['l2'].to_numpy(),
                  index = pd.date_range(start='2000-01-01', periods=len(series['2']), freq='D'),
                  name  ='l2'
              ),
    }

    expected_series_indexes = {
        k: v.index
        for k, v in expected_series_dict.items()
    }
    
    assert isinstance(series_dict, dict)
    assert list(series_dict.keys()) == ['l1', 'l2']
    for k in series_dict:
        assert isinstance(series_dict[k], pd.Series)
        pd.testing.assert_series_equal(series_dict[k], expected_series_dict[k])
        assert isinstance(series_dict[k].index, pd.DatetimeIndex)

    indexes_freq = [f'{v.index.freq}' 
                    for v in series_dict.values()]
    assert len(set(indexes_freq)) == 1

    assert isinstance(series_indexes, dict)
    assert list(series_indexes.keys()) == ['l1', 'l2']
    for k in series_indexes:
        pd.testing.assert_index_equal(series_indexes[k], expected_series_indexes[k])
        assert series_indexes[k].freq == expected_series_indexes[k].freq


# Unit test bayesian_search_forecaster_multiseries
# ==============================================================================
import re
import pytest
import joblib
import warnings
import numpy as np
import pandas as pd
from pathlib import Path
from sklearn.preprocessing import StandardScaler
from lightgbm import LGBMRegressor
from sklearn.linear_model import Ridge
from skforecast.ForecasterAutoregMultiSeries import ForecasterAutoregMultiSeries
from skforecast.ForecasterAutoregMultiSeriesCustom import ForecasterAutoregMultiSeriesCustom
from skforecast.ForecasterAutoregMultiVariate import ForecasterAutoregMultiVariate
from skforecast.model_selection_multiseries import bayesian_search_forecaster_multiseries
from skforecast.model_selection_multiseries import bayesian_search_forecaster_multivariate

# Fixtures
from .fixtures_model_selection_multiseries import series
THIS_DIR = Path(__file__).parent
series_dict = joblib.load(THIS_DIR/'fixture_sample_multi_series.joblib')
exog_dict = joblib.load(THIS_DIR/'fixture_sample_multi_series_exog.joblib')
end_train = "2016-07-31 23:59:00"
series_dict_train = {k: v.loc[:end_train,] for k, v in series_dict.items()}
exog_dict_train = {k: v.loc[:end_train,] for k, v in exog_dict.items()}
series_dict_test = {k: v.loc[end_train:,] for k, v in series_dict.items()}
exog_dict_test = {k: v.loc[end_train:,] for k, v in exog_dict.items()}

def create_predictors(y): # pragma: no cover
    """
    Create first 4 lags of a time series.
    """

    lags = y[-1:-5:-1]

    return lags

def create_predictors_14(y): # pragma: no cover
    """
    Create first 14 lags of a time series.
    """
    lags = y[-1:-15:-1]

    return lags


def test_ValueError_bayesian_search_forecaster_multiseries_when_return_best_and_len_series_exog_different():
    """
    Test ValueError is raised in bayesian_search_forecaster_multiseries when 
    return_best and length of `series` and `exog` do not match.
    """
    forecaster = ForecasterAutoregMultiSeries(
                     regressor = Ridge(random_state=123),
                     lags      = 2,
                     encoding  = 'onehot'
                 )
    exog = series[:30].copy()

    def search_space(trial): # pragma: no cover
        search_space  = {'alpha' : trial.suggest_float('alpha', 1e-2, 1.0)}

        return search_space

    err_msg = re.escape(
        (f"`exog` must have same number of samples as `series`. "
         f"length `exog`: ({len(exog)}), length `series`: ({len(series)})")
    )
    with pytest.raises(ValueError, match = err_msg):
        bayesian_search_forecaster_multiseries(
            forecaster         = forecaster,
            series             = series,
            exog               = exog,
            search_space       = search_space,
            steps              = 3,
            metric             = 'mean_absolute_error',
            refit              = True,
            initial_train_size = len(series[:-12]),
            fixed_train_size   = True,
            n_trials           = 10,
            random_state       = 123,
            return_best        = True,
            verbose            = False,
            engine             = 'optuna'
        )


def test_bayesian_search_forecaster_multiseries_ValueError_when_engine_not_optuna():
    """
    Test ValueError in bayesian_search_forecaster_multiseries is raised when engine 
    is not 'optuna'.
    """
    forecaster = ForecasterAutoregMultiSeries(
                     regressor = Ridge(random_state=123),
                     lags      = 2,
                     encoding  = 'onehot'
                 )

    def search_space(trial): # pragma: no cover
        search_space  = {'alpha': trial.suggest_float('alpha', 1e-2, 1.0)}

        return search_space

    engine = 'not_optuna'
    
    err_msg = re.escape(f"`engine` only allows 'optuna', got {engine}.")
    with pytest.raises(ValueError, match = err_msg):
        bayesian_search_forecaster_multiseries(
            forecaster         = forecaster,
            series             = series,
            search_space       = search_space,
            steps              = 3,
            metric             = 'mean_absolute_error',
            refit              = True,
            initial_train_size = len(series[:-12]),
            fixed_train_size   = True,
            n_trials           = 10,
            random_state       = 123,
            return_best        = False,
            verbose            = False,
            engine             = engine
        )


def test_results_output_bayesian_search_forecaster_multiseries_optuna_engine_ForecasterAutoregMultiSeries():
    """
    Test output of bayesian_search_forecaster_multiseries in 
    ForecasterAutoregMultiSeries with mocked using 
    optuna engine (mocked done in Skforecast v0.12.0).
    """
    forecaster = ForecasterAutoregMultiSeries(
                     regressor = Ridge(random_state=123),
                     lags      = 2,
                     encoding  = 'onehot'
                 )
    steps = 3
    n_validation = 12

    def search_space(trial):
        search_space  = {
            'alpha': trial.suggest_float('alpha', 1e-2, 1.0),
            'lags' : trial.suggest_categorical('lags', [2, 4])
        }

        return search_space

    results = bayesian_search_forecaster_multiseries(
                  forecaster         = forecaster,
                  series             = series,
                  steps              = steps,
                  search_space       = search_space,
                  metric             = 'mean_absolute_error',
                  refit              = False,
                  initial_train_size = len(series) - n_validation,
                  n_trials           = 10,
                  random_state       = 123,
                  return_best        = False,
                  verbose            = False,
                  engine             = 'optuna'
              )[0]
    
    expected_results = pd.DataFrame(
        np.array([[list(['l1', 'l2']), np.array([1, 2, 3, 4]),
            {'alpha': 0.23598059857016607}, 0.20880273566554797,
            0.23598059857016607],
        [list(['l1', 'l2']), np.array([1, 2, 3, 4]),
            {'alpha': 0.398196343012209}, 0.2088108334929226,
            0.398196343012209],
        [list(['l1', 'l2']), np.array([1, 2, 3, 4]),
            {'alpha': 0.4441865222328282}, 0.2088131177203144,
            0.4441865222328282],
        [list(['l1', 'l2']), np.array([1, 2, 3, 4]),
            {'alpha': 0.53623586010342}, 0.2088176743151537,
            0.53623586010342],
        [list(['l1', 'l2']), np.array([1, 2, 3, 4]),
            {'alpha': 0.7252189487445193}, 0.2088269659234972,
            0.7252189487445193],
        [list(['l1', 'l2']), np.array([1, 2]), {'alpha': 0.5558016213920624},
            0.20913532870732662, 0.5558016213920624],
        [list(['l1', 'l2']), np.array([1, 2]), {'alpha': 0.6995044937418831},
            0.20913908163271674, 0.6995044937418831],
        [list(['l1', 'l2']), np.array([1, 2]), {'alpha': 0.7406154516747153},
            0.20914015254956903, 0.7406154516747153],
        [list(['l1', 'l2']), np.array([1, 2]), {'alpha': 0.8509374761370117},
            0.20914302038975385, 0.8509374761370117],
        [list(['l1', 'l2']), np.array([1, 2]), {'alpha': 0.9809565564007693},
            0.20914638910039665, 0.9809565564007693]], dtype=object),
        columns=['levels', 'lags', 'params', 'mean_absolute_error', 'alpha'],
        index=pd.Index([9, 3, 4, 6, 8, 1, 0, 5, 7, 2], dtype='int64')
    ).astype({
        'mean_absolute_error': float, 
        'alpha': float
    })

    pd.testing.assert_frame_equal(results, expected_results)


def test_results_output_bayesian_search_forecaster_multiseries_optuna_engine_ForecasterAutoregMultiSeriesCustom():
    """
    Test output of bayesian_search_forecaster_multiseries in 
    ForecasterAutoregMultiSeriesCustom with mocked using 
    optuna engine (mocked done in Skforecast v0.12.0).
    """
    forecaster = ForecasterAutoregMultiSeriesCustom(
                     regressor      = Ridge(random_state=123),
                     fun_predictors = create_predictors,
                     window_size    = 4,
                     encoding       = 'onehot'
                 )
    steps = 3
    n_validation = 12

    def search_space(trial):
        search_space  = {'alpha': trial.suggest_float('alpha', 1e-2, 1.0)}
        
        return search_space

    results = bayesian_search_forecaster_multiseries(
                  forecaster         = forecaster,
                  series             = series,
                  steps              = steps,
                  search_space       = search_space,
                  metric             = 'mean_absolute_error',
                  refit              = True,
                  initial_train_size = len(series) - n_validation,
                  fixed_train_size   = True,
                  n_trials           = 10,
                  random_state       = 123,
                  return_best        = False,
                  verbose            = False,
                  engine             = 'optuna'
              )[0]
    
    expected_results = pd.DataFrame(
        np.array([[list(['l1', 'l2']), 'custom function: create_predictors',
            {'alpha': 0.2345829390285611}, 0.21476183342122757,
            0.2345829390285611],
        [list(['l1', 'l2']), 'custom function: create_predictors',
            {'alpha': 0.29327794160087567}, 0.21476722307839208,
            0.29327794160087567],
        [list(['l1', 'l2']), 'custom function: create_predictors',
            {'alpha': 0.398196343012209}, 0.21477679099211167,
            0.398196343012209],
        [list(['l1', 'l2']), 'custom function: create_predictors',
            {'alpha': 0.42887539552321635}, 0.2147795727959785,
            0.42887539552321635],
        [list(['l1', 'l2']), 'custom function: create_predictors',
            {'alpha': 0.48612258246951734}, 0.21478474449018078,
            0.48612258246951734],
        [list(['l1', 'l2']), 'custom function: create_predictors',
            {'alpha': 0.5558016213920624}, 0.2147910057902114,
            0.5558016213920624],
        [list(['l1', 'l2']), 'custom function: create_predictors',
            {'alpha': 0.6879814411990146}, 0.21480278321679924,
            0.6879814411990146],
        [list(['l1', 'l2']), 'custom function: create_predictors',
            {'alpha': 0.6995044937418831}, 0.2148038037681593,
            0.6995044937418831],
        [list(['l1', 'l2']), 'custom function: create_predictors',
            {'alpha': 0.7222742800877074}, 0.2148058175049471,
            0.7222742800877074],
        [list(['l1', 'l2']), 'custom function: create_predictors',
            {'alpha': 0.9809565564007693}, 0.2148284279387544,
            0.9809565564007693]], dtype=object),
        columns=['levels', 'lags', 'params', 'mean_absolute_error', 'alpha'],
        index=pd.Index([2, 1, 9, 5, 8, 3, 7, 0, 4, 6], dtype='int64')
    ).astype({
        'mean_absolute_error': float,
        'alpha': float
    })

    pd.testing.assert_frame_equal(results, expected_results)


def test_output_bayesian_search_forecaster_multiseries_series_and_exog_dict_with_mocked():
    """
    Test output of bayesian_search_forecaster_multiseries in ForecasterAutoregMultiSeries 
    and ForecasterAutoregMultiSeriesCustom when series and exog are
    dictionaries (mocked done in Skforecast v0.12.0).
    """

    forecaster = ForecasterAutoregMultiSeries(
        regressor=LGBMRegressor(
            n_estimators=2, random_state=123, verbose=-1, max_depth=2
        ),
        lags=14,
        encoding='ordinal',
        dropna_from_series=False,
        transformer_series=StandardScaler(),
        transformer_exog=StandardScaler(),
    )

    lags_grid = [[5], [1, 7, 14]]

    def search_space(trial):
        search_space = {
            "n_estimators": trial.suggest_int("n_estimators", 2, 5),
            "max_depth": trial.suggest_int("max_depth", 2, 5),
            "lags": trial.suggest_categorical("lags", lags_grid),
        }

        return search_space

    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", category=UserWarning, module="optuna")
        
        results_search, best_trial = bayesian_search_forecaster_multiseries(
            forecaster         = forecaster,
            series             = series_dict,
            exog               = exog_dict,
            search_space       = search_space,
            metric             = 'mean_absolute_error',
            initial_train_size = len(series_dict_train['id_1000']),
            steps              = 24,
            refit              = False,
            n_trials           = 3,
            return_best        = False,
            show_progress      = False,
            verbose            = False,
            suppress_warnings  = True
        )
    
    expected = pd.DataFrame(
        np.array([[list(['id_1000', 'id_1001', 'id_1002', 'id_1003', 'id_1004']),
            np.array([ 1,  7, 14]), {'n_estimators': 4, 'max_depth': 3},
            709.8836514262415, 4, 3],
            [list(['id_1000', 'id_1001', 'id_1002', 'id_1003', 'id_1004']),
            np.array([ 1,  7, 14]), {'n_estimators': 3, 'max_depth': 3},
            721.1848222120482, 3, 3],
            [list(['id_1000', 'id_1001', 'id_1002', 'id_1003', 'id_1004']),
            np.array([5]), {'n_estimators': 4, 'max_depth': 3},
            754.3537196425694, 4, 3]], dtype=object),
        columns=['levels', 'lags', 'params', 'mean_absolute_error', 'n_estimators', 'max_depth'],
        index=pd.Index([0, 2, 1], dtype='int64')
    ).astype({
        'mean_absolute_error': float,
        'n_estimators': int,
        'max_depth': int
    })
    
    results_search = results_search.astype({
        'mean_absolute_error': float,
        'n_estimators': int,
        'max_depth': int
    })

    pd.testing.assert_frame_equal(expected, results_search)


def test_output_bayesian_search_forecaster_multiseries_series_and_exog_dict_with_mocked_ForecasterAutoregMultiSeriesCustom():
    """
    Test output of bayesian_search_forecaster_multiseries in ForecasterAutoregMultiSeriesCustom 
    and ForecasterAutoregMultiSeriesCustom when series and exog are
    dictionaries (mocked done in Skforecast v0.12.0).
    """

    forecaster = ForecasterAutoregMultiSeriesCustom(
        regressor=LGBMRegressor(
            n_estimators=2, random_state=123, verbose=-1, max_depth=2
        ),
        fun_predictors=create_predictors_14, 
        window_size=14,
        encoding='ordinal',
        dropna_from_series=False,
        transformer_series=StandardScaler(),
        transformer_exog=StandardScaler(),
    )

    def search_space(trial):
        search_space = {
            "n_estimators": trial.suggest_int("n_estimators", 2, 5),
            "max_depth": trial.suggest_int("max_depth", 2, 5)
        }

        return search_space

    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", category=UserWarning, module="optuna")
        
        results_search, best_trial = bayesian_search_forecaster_multiseries(
            forecaster         = forecaster,
            series             = series_dict,
            exog               = exog_dict,
            search_space       = search_space,
            metric             = 'mean_absolute_error',
            initial_train_size = len(series_dict_train['id_1000']),
            steps              = 24,
            refit              = False,
            n_trials           = 3,
            return_best        = False,
            show_progress      = False,
            verbose            = False,
            suppress_warnings  = True
        )
    
    expected = pd.DataFrame(
        np.array([[list(['id_1000', 'id_1001', 'id_1002', 'id_1003', 'id_1004']),
            'custom function: create_predictors_14',
            {'n_estimators': 4, 'max_depth': 3}, 709.5984345666709, 4, 3],
            [list(['id_1000', 'id_1001', 'id_1002', 'id_1003', 'id_1004']),
            'custom function: create_predictors_14',
            {'n_estimators': 4, 'max_depth': 3}, 709.5984345666709, 4, 3],
            [list(['id_1000', 'id_1001', 'id_1002', 'id_1003', 'id_1004']),
            'custom function: create_predictors_14',
            {'n_estimators': 2, 'max_depth': 4}, 746.7287691073261, 2, 4]],
            dtype=object),
        columns=['levels', 'lags', 'params', 'mean_absolute_error', 'n_estimators', 'max_depth'],
        index=pd.Index([0, 2, 1], dtype='int64')
    ).astype({
        'mean_absolute_error': float,
        'n_estimators': int,
        'max_depth': int
    })
    
    results_search = results_search.astype({
        'mean_absolute_error': float,
        'n_estimators': int,
        'max_depth': int
    })

    pd.testing.assert_frame_equal(expected, results_search)


def test_results_output_bayesian_search_forecaster_multivariate_optuna_engine_ForecasterAutoregMultiVariate():
    """
    Test output of bayesian_search_forecaster_multivariate in 
    ForecasterAutoregMultiVariate with mocked using 
    optuna engine (mocked done in Skforecast v0.12.0).
    """
    forecaster = ForecasterAutoregMultiVariate(
                     regressor = Ridge(random_state=123),
                     level     = 'l1',
                     lags      = 2,
                     steps     = 3
                 )

    steps = 3
    n_validation = 12

    def search_space(trial):
        search_space  = {
            'alpha': trial.suggest_float('alpha', 1e-2, 1.0),
            'lags' : trial.suggest_categorical('lags', [2, {'l1': 4, 'l2': [2, 3]}])
        }

        return search_space

    results = bayesian_search_forecaster_multivariate(
                  forecaster         = forecaster,
                  series             = series,
                  steps              = steps,
                  search_space       = search_space,
                  metric             = 'mean_absolute_error',
                  refit              = False,
                  initial_train_size = len(series) - n_validation,
                  n_trials           = 10,
                  random_state       = 123,
                  return_best        = False,
                  verbose            = False,
                  engine             = 'optuna'
              )[0]
    
    expected_results = pd.DataFrame(
        np.array([[list(['l1']), {'l1': np.array([1, 2, 3, 4]), 'l2': np.array([2, 3])},
            {'alpha': 0.23598059857016607}, 0.19308110319514993,
            0.23598059857016607],
        [list(['l1']), {'l1': np.array([1, 2, 3, 4]), 'l2': np.array([2, 3])},
            {'alpha': 0.398196343012209}, 0.1931744420708601,
            0.398196343012209],
        [list(['l1']), {'l1': np.array([1, 2, 3, 4]), 'l2': np.array([2, 3])},
            {'alpha': 0.4441865222328282}, 0.1932004954044704,
            0.4441865222328282],
        [list(['l1']), {'l1': np.array([1, 2, 3, 4]), 'l2': np.array([2, 3])},
            {'alpha': 0.53623586010342}, 0.19325210858832276,
            0.53623586010342],
        [list(['l1']), {'l1': np.array([1, 2, 3, 4]), 'l2': np.array([2, 3])},
            {'alpha': 0.7252189487445193}, 0.19335589494249983,
            0.7252189487445193],
        [list(['l1']), np.array([1, 2]), {'alpha': 0.5558016213920624},
            0.20131081099888368, 0.5558016213920624],
        [list(['l1']), np.array([1, 2]), {'alpha': 0.6995044937418831},
            0.2013710017368262, 0.6995044937418831],
        [list(['l1']), np.array([1, 2]), {'alpha': 0.7406154516747153},
            0.2013880862681147, 0.7406154516747153],
        [list(['l1']), np.array([1, 2]), {'alpha': 0.8509374761370117},
            0.20143363961627603, 0.8509374761370117],
        [list(['l1']), np.array([1, 2]), {'alpha': 0.9809565564007693},
            0.20148678375852938, 0.9809565564007693]], dtype=object),
        columns=['levels', 'lags', 'params', 'mean_absolute_error', 'alpha'],
        index=pd.Index([9, 3, 4, 6, 8, 1, 0, 5, 7, 2], dtype='int64')
    ).astype({
        'mean_absolute_error': float,
        'alpha': float
    })

    pd.testing.assert_frame_equal(results, expected_results)